#!/usr/bin/env python

import argparse

from thrift import TSerialization
from thrift.protocol import TJSONProtocol
import unicodecsv

from concrete.util import CommunicationReader


def main():
    parser = argparse.ArgumentParser(
        description="Encode a Communication archive as a CSV file, where each row contains a "
        "TJSONProtocol encoded Communication",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('comms_archive',
                        help="A directory, TGZ file or Zip file of Communications")
    parser.add_argument('csv_file',
                        help="Output CSV file with TJSONProtocol encoded Communications")
    parser.add_argument('--column-name', default='comm',
                        help="Name to use for CSV column header")
    args = parser.parse_args()

    csv_fh = open(args.csv_file, 'wb')

    fieldnames = [args.column_name]
    writer = unicodecsv.DictWriter(csv_fh, fieldnames, lineterminator='\n',
                                   quoting=unicodecsv.QUOTE_ALL)
    writer.writeheader()

    for (comm, filename) in CommunicationReader(args.comms_archive):
        json_communication = TSerialization.serialize(
            comm, TJSONProtocol.TJSONProtocolFactory()).decode('utf-8')
        writer.writerow({args.column_name: json_communication})


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.services.ttypes
import concrete.uuid.ttypes
import concrete.search.ttypes
import concrete.communication.ttypes

from thrift.transport import TTransport
all_structs = []
fix_spec(all_structs)
del all_structs



from __future__ import unicode_literals
from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TCompactProtocol
from thrift.server import TServer


class ThriftFactory(object):
    """Abstract factory to create Thrift objects for client and server."""

    def __init__(self, transportFactory, protocolFactory):
        self.transportFactory = transportFactory
        self.protocolFactory = protocolFactory

    def createSocket(self, host, port):
        '''
        Return new thrift socket.

        Args:
            host (str): hostname to create socket on
            port (int): port number to create socket on

        Returns:
            :class:`TSocket.TSocket`
        '''
        socket = TSocket.TSocket(host, port)
        return socket

    def createTransport(self, socket):
        '''
        Return new thrift transport on socket..

        Args:
            socket (TSocket.TSocket): socket to create transport on

        Returns:
            :class:`TSocket.TSocket`
        '''
        transport = self.transportFactory.getTransport(socket)
        return transport

    def createProtocol(self, transport):
        '''
        Return new thrift protocol on transport.

        Args:
            transport (TTransport.TTransport): transport to create
                protocol on

        Returns:
            :class:`TTransport.TTransport`
        '''
        protocol = self.protocolFactory.getProtocol(transport)
        return protocol

    def createServer(self, processor, host, port):
        '''
        Return new thrift server given a service handler and the
        server host and port.

        Args:
            processor: concrete service handler
            host (str): hostname to serve on
            port (int): port number to serve on

        Returns:
            :class:`TServer.TThreadedServer`
        '''
        socket = TSocket.TServerSocket(host, port)

        server = TServer.TThreadedServer(processor, socket,
                                         self.transportFactory,
                                         self.protocolFactory)
        return server


factory = ThriftFactory(TTransport.TFramedTransportFactory(),
                        TCompactProtocol.TCompactProtocolAcceleratedFactory())


def is_accelerated():
    '''
    Return whether this concrete-python installation has accelerated
    serialization.

    Returns:
        True if this concrete-python installation is accelerated, False
        otherwise
    '''
    try:
        transport = TTransport.TMemoryBuffer()
        TCompactProtocol.TCompactProtocolAccelerated(transport, fallback=False)
        return True
    except Exception:
        return False



#!/usr/bin/env python

'''
Read a concrete tarball and write it back out, rewriting UUIDs with
compressible UUID scheme
'''
from __future__ import unicode_literals

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging

from concrete.util.file_io import CommunicationReader, CommunicationWriterTGZ
from concrete.util.concrete_uuid import compress_uuids as _compress_uuids
from concrete.util import set_stdout_encoding
import concrete.version


def compress_uuids(input_path, output_path, verify=False, uuid_map_path=None,
                   single_analytic=False):
    reader = CommunicationReader(input_path, add_references=False)
    writer = CommunicationWriterTGZ(output_path)

    if uuid_map_path is None:
        uuid_map_file = None
    else:
        uuid_map_file = open(uuid_map_path, 'w')

    for (i, (comm, comm_filename)) in enumerate(reader):
        (new_comm, uc) = _compress_uuids(comm, verify=verify,
                                         single_analytic=single_analytic)

        logging.info('compressed %s (%d analytics, %d uuids) (%d/?)'
                     % (comm.id, len(uc.augs), len(uc.uuid_map), i + 1))

        if uuid_map_file is not None:
            for (old_uuid, new_uuid) in sorted(uc.uuid_map.items(),
                                               key=lambda p: str(p[1])):
                uuid_map_file.write('%s %s\n' % (old_uuid, new_uuid))

        writer.write(new_comm, comm_filename=comm_filename)


def main():
    set_stdout_encoding()

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Read a concrete tarball and write it back out, rewriting'
                    ' UUIDs with compressible UUID scheme',
    )
    parser.set_defaults(log_level='INFO')
    parser.add_argument('input_path', type=str,
                        help='Input tarball path (- for stdin)')
    parser.add_argument('output_path', type=str,
                        help='Output tarball path (- for stdout)')
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    parser.add_argument('--verify', action='store_true',
                        help='Verify within-communication links are satisfied'
                             ' after conversion')
    parser.add_argument('--single-analytic', action='store_true',
                        help='Use one 10-byte prefix for all UUIDs in the'
                             ' communication rather than keying 4 of those'
                             ' bytes on the metadata tool (does not make a'
                             ' noticeable difference in size for typical'
                             ' communications)')
    parser.add_argument('--uuid-map-path', type=str,
                        help='Output path of UUID map')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    # Won't work on Windows
    input_path = '/dev/fd/0' if args.input_path == '-' else args.input_path
    output_path = '/dev/fd/1' if args.output_path == '-' else args.output_path

    compress_uuids(input_path, output_path, verify=args.verify,
                   single_analytic=args.single_analytic,
                   uuid_map_path=args.uuid_map_path)


if __name__ == "__main__":
    main()



from __future__ import unicode_literals
import mock
from mock import sentinel
from pytest import fixture

from concrete.search import SearchProxyService
from concrete.util import (
    SearchProxyServiceWrapper, SearchProxyClientWrapper
)
from concrete.util import ThriftFactory


@fixture
def proxy_client_wrapper_triple():
    host = 'fake-host'
    port = 2
    return (host, port, SearchProxyClientWrapper(host, port))


@fixture
def proxy_service_wrapper():
    class Implementation(SearchProxyService.Iface):
        def search(self, query, provider):
            raise NotImplementedError

        def getProviders(self):
            raise NotImplementedError

        def getCapabilities(self, provider):
            raise NotImplementedError

        def getCorpora(self, provider):
            raise NotImplementedError

        def about(self):
            raise NotImplementedError

        def alive(self):
            raise NotImplementedError

    implementation = Implementation()

    return SearchProxyServiceWrapper(implementation)


@mock.patch('concrete.search.SearchProxyService.Client')
@mock.patch.object(ThriftFactory, 'createProtocol',
                   return_value=sentinel.protocol)
@mock.patch.object(ThriftFactory, 'createTransport')
@mock.patch.object(ThriftFactory, 'createSocket',
                   return_value=sentinel.socket)
def test_enter(mock_create_socket, mock_create_transport,
               mock_create_protocol, mock_client,
               proxy_client_wrapper_triple):
    (host, port, proxy_client_wrapper) = proxy_client_wrapper_triple

    # create additional mocks for transport.open call...
    mock_transport = mock.Mock()
    mock_create_transport.return_value = mock_transport
    # ...and to verify the instantiation of the
    # SearchProxyService.Client
    mock_client.return_value = sentinel.client

    client = proxy_client_wrapper.__enter__()
    # check return value
    assert sentinel.client == client

    # verify method invocations
    mock_create_socket.assert_called_once_with(host, port)
    mock_create_transport.assert_called_once_with(
        mock_create_socket.return_value)
    mock_create_protocol.assert_called_once_with(
        mock_create_transport.return_value)
    mock_client.assert_called_once_with(mock_create_protocol.return_value)

    mock_transport.open.assert_called_once_with()


def test_exit(proxy_client_wrapper_triple):
    (host, port, proxy_client_wrapper) = proxy_client_wrapper_triple

    # create mock for transport.close call
    mock_transport = mock.Mock()
    proxy_client_wrapper.transport = mock_transport

    proxy_client_wrapper.__exit__(mock.ANY, mock.ANY, mock.ANY)

    # verify invocations
    mock_transport.close.assert_called_once_with()


@mock.patch.object(ThriftFactory, 'createServer')
def test_serve(mock_create_server, proxy_service_wrapper):
    # create mock for server.serve invocation
    mock_server = mock.Mock()
    mock_create_server.return_value = mock_server

    host = 'fake-host'
    port = 2

    proxy_service_wrapper.serve(host, port)

    # verify method invocations
    mock_create_server.assert_called_once_with(
        proxy_service_wrapper.processor, host, port)
    mock_server.serve.assert_called_once_with()



#!/usr/bin/env python

"""
Command line script to (partially) validate a Concrete Communication

This script is a thin wrapper around the functionality in the
concrete.validate library.
"""
from __future__ import unicode_literals

import argparse
import logging

import concrete.version
from concrete.validate import validate_communication_file
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description="Validate a Concrete Communication file")
    parser.add_argument('communication_file')
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    validate_communication_file(args.communication_file)


if __name__ == "__main__":
    main()



from __future__ import unicode_literals
from concrete.util import (
    generate_UUID, bin_to_hex, hex_to_bin, AnalyticUUIDGeneratorFactory,
    generate_uuid_unif, generate_hex_unif, split_uuid, join_uuid,
)

from concrete import AnnotationMetadata, Communication
from concrete.validate import validate_communication

import re
import time

from pytest import raises

UUID_RE = re.compile(
    r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$')


def test_generate_UUID():
    comm = Communication()
    comm.uuid = generate_UUID()


def test_validate_minimal_communication_with_uuid():
    comm = Communication()
    comm.id = "myID"
    comm.metadata = AnnotationMetadata(
        tool="TEST", timestamp=int(time.time()))
    comm.type = "Test Communication"
    comm.uuid = generate_UUID()
    assert validate_communication(comm)


def test_hex_to_bin():
    assert hex_to_bin('0') == 0
    assert hex_to_bin('1') == 1
    assert hex_to_bin('2') == 2
    assert hex_to_bin('3') == 3
    assert hex_to_bin('4') == 4
    assert hex_to_bin('5') == 5
    assert hex_to_bin('6') == 6
    assert hex_to_bin('7') == 7
    assert hex_to_bin('8') == 8
    assert hex_to_bin('9') == 9
    assert hex_to_bin('a') == 10
    assert hex_to_bin('A') == 10
    assert hex_to_bin('b') == 11
    assert hex_to_bin('B') == 11
    assert hex_to_bin('c') == 12
    assert hex_to_bin('C') == 12
    assert hex_to_bin('d') == 13
    assert hex_to_bin('D') == 13
    assert hex_to_bin('e') == 14
    assert hex_to_bin('E') == 14
    assert hex_to_bin('f') == 15
    assert hex_to_bin('F') == 15
    assert hex_to_bin('caf') == 15 + 10 * 16 + 12 * 256
    assert hex_to_bin('cafe3') == 3 + 14 * 16 + 15 * 256 + 10 * 4096 + 12 * 65536


def test_bin_to_hex():
    assert bin_to_hex(0) == '0'
    assert bin_to_hex(1) == '1'
    assert bin_to_hex(2) == '2'
    assert bin_to_hex(3) == '3'
    assert bin_to_hex(4) == '4'
    assert bin_to_hex(5) == '5'
    assert bin_to_hex(6) == '6'
    assert bin_to_hex(7) == '7'
    assert bin_to_hex(8) == '8'
    assert bin_to_hex(9) == '9'
    assert bin_to_hex(10).lower() == 'a'
    assert bin_to_hex(11).lower() == 'b'
    assert bin_to_hex(12).lower() == 'c'
    assert bin_to_hex(13).lower() == 'd'
    assert bin_to_hex(14).lower() == 'e'
    assert bin_to_hex(15).lower() == 'f'
    assert bin_to_hex(15 + 10 * 16 + 12 * 256).lower() == 'caf'
    assert bin_to_hex(3 + 14 * 16 + 15 * 256 + 10 * 4096 + 12 * 65536).lower() == 'cafe3'


def test_bin_to_hex_with_len():
    assert bin_to_hex(0, 4) == '0000'
    assert bin_to_hex(1, 4) == '0001'
    assert bin_to_hex(2, 4) == '0002'
    assert bin_to_hex(3, 4) == '0003'
    assert bin_to_hex(4, 4) == '0004'
    assert bin_to_hex(5, 4) == '0005'
    assert bin_to_hex(6, 4) == '0006'
    assert bin_to_hex(7, 4) == '0007'
    assert bin_to_hex(8, 4) == '0008'
    assert bin_to_hex(9, 4) == '0009'
    assert bin_to_hex(10, 4).lower() == '000a'
    assert bin_to_hex(11, 4).lower() == '000b'
    assert bin_to_hex(12, 4).lower() == '000c'
    assert bin_to_hex(13, 4).lower() == '000d'
    assert bin_to_hex(14, 4).lower() == '000e'
    assert bin_to_hex(15, 4).lower() == '000f'
    assert bin_to_hex(15 + 10 * 16 + 12 * 256, 4).lower() == '0caf'
    with raises(ValueError):
        bin_to_hex(3 + 14 * 16 + 15 * 256 + 10 * 4096 + 12 * 65536, 4)


def test_split_uuid_valid():
    assert (
        split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e16') ==
        ('7575a428aaf7', '4c2e929e', '1e2a0ab59e16')
    )


def test_split_uuid_too_many_pieces():
    with raises(ValueError):
        split_uuid('aaf7-7575a428-aaf7-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_too_few_pieces():
    with raises(ValueError):
        split_uuid('7575a428-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_first_piece_too_long():
    with raises(ValueError):
        split_uuid('7575a428a-aaf7-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_second_piece_too_long():
    with raises(ValueError):
        split_uuid('7575a428-aaf7a-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_third_piece_too_long():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2ea-929e-1e2a0ab59e16')


def test_split_uuid_fourth_piece_too_long():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2e-929ea-1e2a0ab59e16')


def test_split_uuid_fifth_piece_too_long():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e16a')


def test_split_uuid_first_piece_too_short():
    with raises(ValueError):
        split_uuid('7575a42-aaf7-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_second_piece_too_short():
    with raises(ValueError):
        split_uuid('7575a428-aaf-4c2e-929e-1e2a0ab59e16')


def test_split_uuid_third_piece_too_short():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2-929e-1e2a0ab59e16')


def test_split_uuid_fourth_piece_too_short():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2e-929-1e2a0ab59e16')


def test_split_uuid_fifth_piece_too_short():
    with raises(ValueError):
        split_uuid('7575a428-aaf7-4c2e-929e-1e2a0ab59e1')


def test_join_uuid_valid():
    assert (
        join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e16') ==
        '7575a428-aaf7-4c2e-929e-1e2a0ab59e16'
    )


def test_join_uuid_first_piece_too_long():
    with raises(ValueError):
        join_uuid('7575a428aaf7a', '4c2e929e', '1e2a0ab59e16')


def test_join_uuid_second_piece_too_long():
    with raises(ValueError):
        join_uuid('7575a428aaf7', '4c2e929ea', '1e2a0ab59e16')


def test_join_uuid_third_piece_too_long():
    with raises(ValueError):
        join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e16a')


def test_join_uuid_first_piece_too_short():
    with raises(ValueError):
        join_uuid('7575a428aaf', '4c2e929e', '1e2a0ab59e16')


def test_join_uuid_second_piece_too_short():
    with raises(ValueError):
        join_uuid('7575a428aaf7', '4c2e929', '1e2a0ab59e16')


def test_join_uuid_third_piece_too_short():
    with raises(ValueError):
        join_uuid('7575a428aaf7', '4c2e929e', '1e2a0ab59e1')


def test_generate_hex_unif_range():
    n = 1000  # (1 - 1/16)^n = 9e-29
    r = set()
    for i in range(n):
        r.add(generate_hex_unif(1))
    assert (
        sorted(map(lambda x: str(x).lower(), r)) ==
        [c for c in '0123456789abcdef']
    )


def test_generate_hex_unif_len():
    h = generate_hex_unif(21)  # (1/16)^(n-1) = 8e-25
    assert len(h) == 21
    assert len(set(c for c in h)) > 1


def test_generate_hex_unif_spread():
    n = 1000
    m = 32
    # union bound: (1/16)^m * n^2 = 3e-33
    s = set([generate_hex_unif(m) for i in range(n)])
    assert len(s) == n


def test_generate_uuid_unif_format():
    u = generate_uuid_unif()
    assert UUID_RE.match(u) is not None


def test_generate_uuid_unif_spread():
    n = 1000
    # union bound: (1/16)^32 * n^2 = 3e-33
    s = set([generate_uuid_unif() for i in range(n)])
    assert len(s) == n


class Duck(object):
    pass


def test_AnalyticUUIDGeneratorFactory_x_prefix_no_comm():
    n = 1000
    augf = AnalyticUUIDGeneratorFactory()
    u = augf.comm_uuid
    for i in range(n):
        aug = augf.create()
        assert next(aug).uuidString.startswith(u[:8 + 1 + 4])


def test_AnalyticUUIDGeneratorFactory_x_prefix_with_comm():
    n = 1000
    u = '7575a428-aaf7-4c2e-929e-1e2a0ab59e16'
    comm = Duck()
    comm.uuid = Duck()
    comm.uuid.uuidString = u
    augf = AnalyticUUIDGeneratorFactory(comm)
    assert augf.comm_uuid == u
    for i in range(n):
        aug = augf.create()
        assert next(aug).uuidString.startswith(u[:8 + 1 + 4])


def test_AnalyticUUIDGeneratorFactory_x_prefix_bad_comm_uuid():
    u = '7575a428a-aaf7-4c2e-929e-1e2a0ab59e16'
    comm = Duck()
    comm.uuid = Duck()
    comm.uuid.uuidString = u
    augf = AnalyticUUIDGeneratorFactory(comm)
    with raises(ValueError):
        augf.create()


def test_AnalyticUUIDGeneratorFactory_y_prefix():
    m = 100
    n = 100
    augf = AnalyticUUIDGeneratorFactory()
    for i in range(m):
        aug = augf.create()
        uu = next(aug).uuidString
        for j in range(n - 1):
            assert next(aug).uuidString.startswith(
                uu[:8 + 1 + 4 + 1 + 4 + 1 + 4])


def test_AnalyticUUIDGeneratorFactory_z_increment():
    m = 100
    n = 100
    augf = AnalyticUUIDGeneratorFactory()
    for i in range(m):
        aug = augf.create()
        u = next(aug).uuidString
        z = int(u[8 + 1 + 4 + 1 + 4 + 1 + 4 + 1:], 16)
        for j in range(n - 1):
            u = next(aug).uuidString
            assert (
                int(u[8 + 1 + 4 + 1 + 4 + 1 + 4 + 1:], 16) ==
                (z + 1) % 2**48
            )
            z = (z + 1) % 2**48


def test_AnalyticUUIDGeneratorFactory_x_prefix_spread():
    m = 100
    # union bound: (1/16)^12 * m^2 = 4e-11
    s = set()
    for i in range(m):
        augf = AnalyticUUIDGeneratorFactory()
        aug = augf.create()
        u = next(aug).uuidString
        s.add(u[:8 + 1 + 4])
    assert len(s) == m


def test_AnalyticUUIDGeneratorFactory_y_prefix_spread():
    m = 10
    augf = AnalyticUUIDGeneratorFactory()
    # union bound: (1/16)^8 * m^2 = 2e-8
    s = set()
    for i in range(m):
        aug = augf.create()
        u = next(aug).uuidString
        s.add(u[:8 + 1 + 4 + 1 + 4 + 1 + 4])
    assert len(s) == m


def test_AnalyticUUIDGeneratorFactory_spread():
    m = 100
    n = 100
    augf = AnalyticUUIDGeneratorFactory()
    # union bound: (2m-1)*(1/16)^12 * n^2 = 7e-9
    s = set()
    for i in range(m):
        aug = augf.create()
        u = next(aug).uuidString
        s.add(u)
        for j in range(n - 1):
            u = next(aug).uuidString
            s.add(u)
    assert len(s) == m * n



"""Utility code for working with Concrete
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from .access import *  # noqa
from .access_wrapper import *  # noqa
from .annotate_wrapper import *  # noqa
from .comm_container import *  # noqa
from .concrete_uuid import *  # noqa
from .file_io import *  # noqa
from .json_fu import *  # noqa
from .learn_wrapper import *  # noqa
from .locale import *  # noqa
from .mem_io import *  # noqa
from .metadata import *  # noqa
from .net import *  # noqa
from .redis_io import *  # noqa
from .references import *  # noqa
from .results_wrapper import *  # noqa
from .search_wrapper import *  # noqa
from .service_wrapper import *  # noqa
from .simple_comm import *  # noqa
from .summarization_wrapper import * # noqa
from .thrift_factory import *  # noqa
from .tokenization import *  # noqa
from .twitter import *  # noqa
from .unnone import *  # noqa



from __future__ import unicode_literals
import logging

from ..structure.ttypes import TokenizationKind
from .unnone import lun

from collections import deque
from math import log, exp
from functools import reduce


class NoSuchTokenTagging(Exception):
    '''
    Exception representing there is no :class:`TokenTagging` annotation
    that matches the given criteria in a given concrete object
    '''
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


def get_tokens(tokenization, suppress_warnings=False):
    """Get list of :class:`.Token` objects for a :class:`.Tokenization`

    Return list of Tokens from lattice.cachedBestPath, if Tokenization
    kind is TOKEN_LATTICE; else, return list of Tokens from tokenList.

    Warn and return list of Tokens from tokenList if kind is not set.

    Return None if kind is set but the respective data fields are not.

    Args:
        tokenization (Tokenization): tokenization to extract tokens
            from
        suppress_warnings (bool): True to suppress warning messages
            that `tokenization.kind` is None

    Returns:
        List of :class:`.Token` objects, or `None`

    Raises:
        ValueError: if `tokenization.kind` is not a recognized
            tokenization kind
    """

    if tokenization.kind is None:
        if not suppress_warnings:
            logging.warning('Tokenization.kind is None but is required in '
                            'Concrete; using tokenList')

        token_list = tokenization.tokenList
        if token_list.tokenList is not None:
            return token_list.tokenList

    elif tokenization.kind == TokenizationKind.TOKEN_LATTICE:
        token_lattice = tokenization.lattice
        if token_lattice.cachedBestPath is not None:
            lattice_path = token_lattice.cachedBestPath
            if lattice_path.tokenList is not None:
                return lattice_path.tokenList

    elif tokenization.kind == TokenizationKind.TOKEN_LIST:
        token_list = tokenization.tokenList
        if token_list.tokenList is not None:
            return token_list.tokenList

    else:
        raise ValueError('Unrecognized TokenizationKind %s'
                         % tokenization.kind)

    return None


def get_token_taggings(tokenization, tagging_type, case_sensitive=False):

    """Return list of :class:`.TokenTagging` objects of `taggingType`
    equal to `tagging_type`.

    Args:
        tokenization (Tokenization): tokenization from which taggings
            will be selected
        tagging_type (str): value of `taggingType` to filter to
        case_sensitive (bool): True to do case-sensitive matching
            on `taggingType`.

    Returns:
        List of :class:`.TokenTagging` objects of `taggingType` equal
        to `tagging_type`, in same order as they appeared in the
        tokenization.  If no matching :class:`.TokenTagging` objects
        exist, return an empty list.
    """
    if not tokenization.tokenTaggingList:
        return []
    else:
        return [
            tt for tt in tokenization.tokenTaggingList
            if (
                (tt.taggingType == tagging_type)
                if case_sensitive else
                (tt.taggingType.lower() == tagging_type.lower())
            )
        ]


def get_tagged_tokens(tokenization, tagging_type, tool=None):

    """Return list of :class:`.TaggedToken` objects of taggingType equal
    to tagging_type, if there is a unique choice.

    Args:
        tokenization (Tokenization): tokenization to return
            tagged tokens for
        tagging_type (str): only return tagged tokens for
            :class:`TokenTagging` objects whose `taggingType`
            field is equal to `tagging_type`
        tool (str): If not None, only return tagged tokens for
            :class:`TokenTagging` objects whose `metadata.tool`
            field is equal to `tool`

    Returns:
        List of :class:`.TaggedToken` objects of `taggingType` equal
        to `tagging_type`, if there is a unique choice.

    Raises:
        NoSuchTokenTagging: if there is no matching tagging
        Exception: if there is more than one matching tagging.
    """
    tts = [
        tt
        for tt in get_token_taggings(tokenization, tagging_type)
        if tool is None or tt.metadata.tool.lower() == tool.lower()
    ]
    if len(tts) == 0:
        raise NoSuchTokenTagging('No matching %s tagging.' % tagging_type)
    elif len(tts) == 1:
        return tts[0].taggedTokenList
    else:
        raise Exception('More than one matching %s tagging.' % tagging_type)


def get_lemmas(t, tool=None):
    """
    Returns the result of :func:`get_tagged_tokens` with a
    `tagging_type` of "LEMMA"

    Args:
        t (Tokenization): tokenization to extract tagged tokens from
        tool (str): If not None, only return tagged tokens for
            :class:`TokenTagging` objects whose `metadata.tool`
            field is equal to `tool`

    Returns:
        list of 'LEMMA'-tagged tokens matching `tool` (if specified)
    """
    return get_tagged_tokens(t, 'LEMMA', tool=tool)


def get_pos(t, tool=None):
    """
    Returns the result of :func:`get_tagged_tokens` with a
    `tagging_type` of "LEMMA"

    Args:
        t (Tokenization): tokenization to extract tagged tokens from
        tool (str): If not None, only return tagged tokens for
            :class:`TokenTagging` objects whose `metadata.tool`
            field is equal to `tool`

    Returns:
        list of 'POS'-tagged tokens matching `tool` (if specified)
    """
    return get_tagged_tokens(t, 'POS', tool=tool)


def get_ner(t, tool=None):
    """
    Returns the result of :func:`get_tagged_tokens` with a
    `tagging_type` of "NER"

    Args:
        t (Tokenization): tokenization to extract tagged tokens from
        tool (str): If not None, only return tagged tokens for
            :class:`TokenTagging` objects whose `metadata.tool`
            field is equal to `tool`

    Returns:
        list of 'NER'-tagged tokens matching `tool` (if specified)
    """
    return get_tagged_tokens(t, 'NER', tool=tool)


def plus(x, y):
    """
    Return concatenation of two lists.

    Args:
        x (list):
        y (list)

    Returns:
        list concatenation of x and y
    """
    return x + y


def flatten(a):
    """
    Returned flattened version of input list.

    Args:
        a (list):
    Returns:
        list: Flattened list
    """
    return reduce(plus, a, [])


def get_comm_tokens(comm, sect_pred=None, suppress_warnings=False):
    """Get list of :class:`.Token` objects in :class:`.Communication`.

    Args:
        comm (Communication): communications to extract tokens from
        sect_pred (function): Function that takes a :class:`.Section`
            and returns false if the :class:`.Section` should be
            excluded.
        suppress_warnings (bool): True to suppress warning messages
            that `Tokenization.kind` is None

    Returns:
        List of :class:`.Token` objects in :class:`.Communication`,
        delegating to :func:`get_tokens` for each sentence.
    """
    return flatten(map(
        lambda sect: flatten(map(
            lambda sent: get_tokens(sent.tokenization, suppress_warnings),
            sect.sentenceList
        )),
        filter(sect_pred, comm.sectionList)
        if (sect_pred is not None) else comm.sectionList
    ))


def get_comm_tokenizations(comm, tool=None):
    """Get list of :class:`.Tokenization` objects in a :class:`.Communication`

    Args:
        comm (Communication): communications to extract tokenizations
            from
        tool (str): If not None, only return :class:`.Tokenization`
            objects whose `metadata.tool` field is equal to `tool`

    Returns:
        List of :class:`.Tokenization` objects
    """
    for section in lun(comm.sectionList):
        for sentence in lun(section.sentenceList):
            if tool is None or sentence.tokenization.metadata.tool == tool:
                yield sentence.tokenization


def _lattice_to_fsm(lattice):
    '''
    Return FSM representation of token lattice.

    Args:
        lattice (TokenLattice): the token lattice to process

    Returns:
        tuple containing four items: 0. arcs represented as
        a dictionary from source states to dictionaries from
        destination states to lists of (token, weight) pairs;
        1. arcs represented as a dictionary from destination states to
        dictionaries from source states to lists of (token, weight)
        pairs; 2. the set of tokens; and 3. the set of states.
    '''
    fsm = {}
    bkFsm = {}

    tokens = set()
    states = set()
    for arc in lattice.arcList:
        if arc.src is None:
            raise ValueError('Arc.src must be set')
        src = arc.src

        if arc.dst is None:
            raise ValueError('Arc.dst must be set')
        dst = arc.dst

        if arc.token is None:
            raise ValueError('Arc.token must be set')
        token = arc.token.tokenIndex

        if arc.weight is None:
            raise ValueError('Arc.weight must be set')
        wt = arc.weight

        tokens.add(token)
        states.add(src)
        states.add(dst)

        if src not in fsm:
            fsm[src] = {}
        if dst not in fsm[src]:
            fsm[src][dst] = []
        fsm[src][dst].append((token, wt))

        if dst not in bkFsm:
            bkFsm[dst] = {}
        if src not in bkFsm[dst]:
            bkFsm[dst][src] = []
        bkFsm[dst][src].append((token, wt))

    return (fsm, bkFsm, tokens, states)


def _logsumexp(a):
    '''
    Return log of sum of exponentiations of elements in a.

    Args:
        a: list or vector (of numbers)

    Returns:
        log of sum of exponentiations of elements in a.
    '''
    try:
        from scipy.misc import logsumexp
        return logsumexp(a)
    except ImportError:
        m = max(a)
        return m + log(sum(map(lambda x: exp(x - m), a)))


def _calc_marginal_in_log_prob(fsm, states, start, end):
    """
    Calculate marginal in-log-probability of each state.
    """

    alpha = {}
    alpha[start] = 0.

    state_queue = deque([start])

    while state_queue:
        currState = state_queue.popleft()

        if currState == end:
            continue

        arcs = fsm[currState]

        for (dst, tokenWts) in arcs.items():
            if dst not in alpha:
                state_queue.append(dst)
                alpha[dst] = float('-inf')
            for token, wt in tokenWts:
                alpha[dst] = _logsumexp([alpha[dst], alpha[currState] + wt])

    return alpha


def compute_lattice_expected_counts(lattice):
    """Given a :class:`.TokenLattice` in which the dst, src, token, and
    weight fields are set in each arc, compute and return a list of
    expected token log-probabilities.

    Input arc weights are treated as unnormalized log-probabilities.

    Args:
        lattice (TokenLattice): lattice to compute expected counts for

    Returns:
        List of floats (expected log-probabilities) with the float
        at position i corresponding to the token with tokenIndex i.
    """

    (fsm, bkFsm, tokens, states) = _lattice_to_fsm(lattice)
    alpha = _calc_marginal_in_log_prob(fsm, states,
                                       lattice.startState, lattice.endState)
    beta = _calc_marginal_in_log_prob(bkFsm, states,
                                      lattice.endState, lattice.startState)
    norm = alpha[lattice.endState]

    expectedCounts = {}

    for state in states:
        currState = state

        if currState == lattice.endState:
            continue

        arcs = fsm[currState]

        for dst, tokenWts in arcs.items():
            for token, wt in tokenWts:
                if token not in expectedCounts:
                    expectedCounts[token] = []
                expectedCounts[token].append(
                    alpha[currState] + beta[dst] + wt - norm)

    if expectedCounts:
        return [
            (_logsumexp(expectedCounts[idx]) if idx in expectedCounts else None)
            for idx in range(max(expectedCounts) + 1)
        ]
    else:
        return []


def get_tokenizations(comm, tool=None):
    """
    Returns a flat list of all Tokenization objects in a Communication

    Args:
        comm (Communication): communication to get tokenizations from
        tool (str): if not None, return only tokenizations whose
            `metadata.tool` field matches `tool`

    Returns:
        A list of all Tokenization objects within the Communication
        matching `tool` (if it is not None)
    """
    tokenizations = []

    if comm.sectionList:
        for section in comm.sectionList:
            if section.sentenceList:
                for sentence in section.sentenceList:
                    if sentence.tokenization:
                        if (tool is None or
                                sentence.tokenization.metadata.tool == tool):
                            tokenizations.append(sentence.tokenization)
    return tokenizations



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import concrete.context.ttypes
import concrete.metadata.ttypes
import concrete.language.ttypes
import concrete.structure.ttypes
import concrete.entities.ttypes
import concrete.situations.ttypes
import concrete.exceptions.ttypes
import concrete.email.ttypes
import concrete.twitter.ttypes
import concrete.audio.ttypes
import concrete.communication.ttypes
import concrete.services.ttypes

from thrift.transport import TTransport
all_structs = []
fix_spec(all_structs)
del all_structs



#!/usr/bin/env python

'''
Convert Tweet file to Concrete Communications file.
'''
from __future__ import unicode_literals

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from multiprocessing import Pool
import gzip
import json
import logging
import logging.config
import mimetypes

import concrete.version
from concrete.validate import validate_communication
from concrete.util.mem_io import write_communication_to_buffer
from concrete.util.twitter import json_tweet_string_to_Communication
from concrete.util import set_stdout_encoding

try:
    JSONDecodeError = json.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError


def json_str_to_validated_concrete_bytes(tweet_str):
    tweet_str = tweet_str.decode('utf-8')
    b = json_tweet_string_to_Communication(tweet_str, True, True)
    if b is None or not validate_communication(b):
        return None
    else:
        return write_communication_to_buffer(b)


def json_str_to_concrete_bytes(tweet_str):
    tweet_str = tweet_str.decode('utf-8')
    b = json_tweet_string_to_Communication(tweet_str, True, True)
    if b is None:
        return None
    else:
        return write_communication_to_buffer(b)


def json_str_to_concrete_bytes_skip_bad_lines(tweet_str):
    try:
        return json_str_to_concrete_bytes(tweet_str)
    except UnicodeDecodeError:
        return None
    except JSONDecodeError:
        return None


def json_str_to_validated_concrete_bytes_skip_bad_lines(tweet_str):
    try:
        return json_str_to_validated_concrete_bytes(tweet_str)
    except UnicodeDecodeError:
        return None
    except JSONDecodeError:
        return None


def main():
    set_stdout_encoding()

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Read tweets formatted in the Twitter JSON API and write'
                    ' communications',
    )
    parser.set_defaults(num_proc=1, chunk_size=10000, log_interval=10000)
    parser.add_argument('tweet_path', type=str,
                        help='Input twitter JSON file path (- for stdin)')
    parser.add_argument('concrete_path', type=str,
                        help='Output concrete file path (- for stdout)')
    parser.add_argument('--log-interval', type=int,
                        help='log an info message every log-interval tweets')
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    parser.add_argument('--log-conf-path', type=str,
                        help='Path to log config file (overrides log-level).'
                             ' Format is json encoding of logging.config'
                             ' dictConfig schema: https://docs.python.org/2/'
                             'library/logging.config.html#'
                             'logging-config-dictschema')
    parser.add_argument('--num-proc', type=int,
                        help='Number of worker processes to use')
    parser.add_argument('--chunk-size', type=int,
                        help='Chunk size (in number of tweets) when'
                             ' dispatching jobs to workers')
    parser.add_argument('--catch-ioerror', action='store_true',
                        help='Treat IOError as successful end of stream'
                             ' (recommended if gzipped input files were not'
                             ' completely written)')
    parser.add_argument('--skip-bad-lines', action='store_true',
                        help='Skip malformatted json lines'
                             ' (by default, they crash the program)')
    parser.add_argument('--skip-invalid-comms', action='store_true',
                        help='Skip invalid communications (increases runtime)'
                             ' (by default, validation is not performed)')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    # Won't work on Windows
    tweet_path = '/dev/fd/0' if args.tweet_path == '-' else args.tweet_path
    concrete_path = (
        '/dev/fd/1' if args.concrete_path == '-' else args.concrete_path
    )

    if args.log_conf_path:
        with open(args.log_conf_path) as f:
            logging.config.dictConfig(json.load(f))
    else:
        logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                            level=args.loglevel.upper())

    if args.tweet_path != '-' and mimetypes.guess_type(tweet_path)[1] == 'gzip':
        tweet_reader = gzip.open(tweet_path, 'rb')
    else:
        tweet_reader = open(tweet_path, 'rb')

    if args.catch_ioerror:
        def _catch(g):
            it = iter(g)
            while True:
                try:
                    x = next(it)
                    yield x
                except IOError:
                    return
                except EOFError:
                    return
        tweet_reader = _catch(tweet_reader)

    if args.skip_bad_lines:
        if args.skip_invalid_comms:
            map_func = json_str_to_validated_concrete_bytes_skip_bad_lines
        else:
            map_func = json_str_to_concrete_bytes_skip_bad_lines
    else:
        if args.skip_invalid_comms:
            map_func = json_str_to_validated_concrete_bytes
        else:
            map_func = json_str_to_concrete_bytes

    with open(concrete_path, 'wb') as writer:
        i = 0

        p = Pool(args.num_proc)
        for concrete_bytes in p.imap(map_func, tweet_reader,
                                     args.chunk_size):
            if concrete_bytes is not None:
                if (i + 1) % args.log_interval == 0:
                    logging.info('writing tweet %d...' % (i + 1))
                writer.write(concrete_bytes)
                i += 1


if __name__ == "__main__":
    main()



#!/usr/bin/env python

from __future__ import print_function
from __future__ import unicode_literals

from time import sleep, time
import argparse
import logging
import sys

import concrete.version
import concrete.util
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = argparse.ArgumentParser(
        description='Wait for a service to come up.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('service_name',
                        choices=[
                            'FetchCommunication',
                            'StoreCommunication',
                            'AnnotateCommunication',
                            'ActiveLearnerClient',
                            'ActiveLearnerServer',
                            'ResultsServer',
                            'Search',
                            'SearchProxy',
                            'Feedback',
                            'Summarization',
                        ],
                        help='name of service to wait for')
    parser.add_argument('--host', help='hostname to serve on',
                        default='localhost')
    parser.add_argument('--port', type=int, help='port to serve on',
                        default=9090)
    parser.add_argument('--sleep-interval', type=int, default=5)
    parser.add_argument('--timeout', type=int,
                        help='Default: no timeout')
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    cli_wrapper_class = getattr(concrete.util,
                                args.service_name + 'ClientWrapper')

    logging.info('waiting for {} service at {}:{} to come up'.format(
        args.service_name, args.host, args.port))
    start = time()
    alive = False
    while not alive:
        try:
            with cli_wrapper_class(args.host, args.port) as cli:
                if cli.alive():
                    alive = True
                    break
        except Exception:
            pass
        if args.timeout is not None and time() - start > args.timeout:
            logging.info('timed out after {}s'.format(args.timeout))
            break
        else:
            logging.info('waiting {}s for {} service to come up'.format(
                args.sleep_interval, args.service_name))
            sleep(args.sleep_interval)

    if alive:
        logging.info('{} service at {}:{} is up'.format(
            args.service_name, args.host, args.port))
    else:
        logging.error('{} service at {}:{} is down'.format(
            args.service_name, args.host, args.port))
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import unicode_literals
from multiprocessing import Process
from time import sleep
from socket import create_connection

from thrift.protocol import TJSONProtocol
from thrift.transport import THttpClient

from .thrift_factory import factory


class ConcreteServiceClientWrapper(object):
    '''
    Base class for a wrapper around a Concrete service client.
    Implements the context manager interface so client can be controlled
    using the `with:` statement (client connection is closed when the
    `with:` scope is exited).
    '''
    def __init__(self, host, port):
        '''
        Args:
            host (str): hostname to connect to
            port (int): port number to connect to
        '''
        if not hasattr(self, 'concrete_service_class'):
            raise NotImplementedError(
                "Child classes of ConcreteServiceClientWrapper must set " +
                "the 'concrete_service_class' attribute to a class that " +
                "implements a Concrete Service")

        self.host = host
        try:
            self.port = int(port)
        except ValueError:
            raise ValueError(
                "Service expected 'port' to be an integer, but it was '%s'" %
                port)

    def __enter__(self):
        '''
        Create and open connection.
        '''
        socket = factory.createSocket(self.host, self.port)
        self.transport = factory.createTransport(socket)
        protocol = factory.createProtocol(self.transport)

        cli = self.concrete_service_class.Client(protocol)

        self.transport.open()
        return cli

    def __exit__(self, type, value, traceback):
        '''
        Close connection.
        '''
        self.transport.close()


class ConcreteServiceWrapper(object):
    """
    Base class for a wrapper around a Concrete service that runs in
    (blocks) the current process.
    """

    def __init__(self, implementation):
        '''
        Args:
            implementation (object): handler of specified concrete
                service
        '''
        if not hasattr(self, 'concrete_service_class'):
            raise NotImplementedError(
                "Child classes of ConcreteServiceWrapper must set " +
                "the 'concrete_service_class' attribute to a class that " +
                "implements a Concrete Service")

        self.processor = self.concrete_service_class.Processor(implementation)

    def serve(self, host, port):
        '''
        Serve on specified host and port in current process, blocking
        until server is killed.  (If server is not killed by signal or
        otherwise it will block forever.)

        Args:
            host (str): hostname to serve on
            port (int): port number to serve on
        '''
        server = factory.createServer(self.processor, host, port)

        server.serve()


class HTTPConcreteServiceClientWrapper(object):
    '''
    Base class for a wrapper around an HTTP Concrete service client.
    Implements the context manager interface so client can be controlled
    using the `with:` statement (client connection is closed when the
    `with:` scope is exited).
    '''
    def __init__(self, uri):
        '''
        Args:
            uri (str):
        '''
        if not hasattr(self, 'concrete_service_class'):
            raise NotImplementedError(
                "Child classes of HTTPConcreteServiceClientWrapper must set " +
                "the 'concrete_service_class' attribute to a class that " +
                "implements a Concrete Service")

        self.uri = uri

    def __enter__(self):
        '''
        Create and open connection
        '''
        self.transport = THttpClient.THttpClient(uri_or_host=self.uri)
        protocol = TJSONProtocol.TJSONProtocol(self.transport)
        cli = self.concrete_service_class.Client(protocol)
        return cli

    def __exit__(self, type, value, traceback):
        '''
        Close connection.
        '''
        self.transport.close()


class SubprocessConcreteServiceWrapper(object):
    """
    Base class for a wrapper around a Concrete service that runs in
    a subprocess; implements the context manager interface so subprocess
    can be controlled using the `with:` statement (subprocess is stopped
    and joined when the `with:` scope is exited).
    """

    SLEEP_INTERVAL = 0.1

    def __init__(self, implementation, host, port, timeout=None):
        '''
        Args:
            implementation (object): handler of specified concrete
                service
            host (str): hostname that will be served on
                when context is entered
            port (int): port number that will be served on
                when context is entered
            timeout (int): number of seconds to wait for server to start
                in subprocess, when context is entered (if None, wait
                forever)
        '''
        if not hasattr(self, 'concrete_service_wrapper_class'):
            raise NotImplementedError(
                "Child classes of SubprocessConcreteServiceWrapper must " +
                "set the 'concrete_service_wrapper_class' attribute to a " +
                "child class of ConcreteServiceWrapper")

        self.proc = None
        self.server = self.concrete_service_wrapper_class(implementation)
        self.host = host
        self.port = port
        self.timeout = timeout

    def __enter__(self):
        '''
        Create and start subprocess.
        '''
        self.proc = Process(target=self.server.serve,
                            args=(self.host, self.port))
        self.proc.start()

        s = None
        elapsed = 0.

        # Loop until successful connect or timeout
        while s is None and (self.timeout is None or elapsed < self.timeout):
            try:
                s = create_connection((self.host, self.port))
                s.close()
            except Exception:
                s = None

                sleep(self.SLEEP_INTERVAL)
                elapsed += self.SLEEP_INTERVAL

        return self

    def __exit__(self, type, value, traceback):
        '''
        Stop and join subprocess.
        '''
        self.proc.terminate()
        self.proc.join()



from __future__ import unicode_literals
import socket


def find_port():
    '''
    Find and return an available TCP port.

    Returns:
        an unused TCP port (an integer)
    '''
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    addr = s.getsockname()
    port = addr[1]
    s.close()
    return port



# -*- coding: utf-8 -*-


from __future__ import unicode_literals
from pytest import fixture, mark
from concrete.util import CommunicationReader
from concrete.validate import validate_communication
import sys
import json
from subprocess import Popen, PIPE


def assert_first_comm(comm):
    assert comm.id == '238426131689242624'
    assert comm.startTime == 1345680194
    assert comm.endTime == 1345680194
    assert validate_communication(comm)


def assert_second_comm(comm):
    assert comm.id == '238426131689242625'
    assert comm.startTime == 1345680195
    assert comm.endTime == 1345680195
    assert validate_communication(comm)


@fixture
def log_conf(tmpdir):
    log_path = str(tmpdir / 'log')
    conf_path = str(tmpdir / 'log.conf')

    with open(conf_path, 'w') as f:
        json.dump(
            dict(
                version=1,
                root=dict(
                    level='INFO',
                    handlers=['file'],
                ),
                formatters=dict(
                    medium=dict(
                        format='%(asctime)-15s %(levelname)s: %(message)s',
                    ),
                ),
                handlers=dict(
                    file={
                        'class': 'logging.FileHandler',
                        'formatter': 'medium',
                        'filename': log_path,
                    },
                ),
            ),
            f
        )

    yield (conf_path, log_path)


@fixture
def output_file(tmpdir):
    yield str(tmpdir / 'output.comm')


def test_tweets2concrete(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


@mark.posix
def test_tweets2concrete_stdin(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '-',
        output_file
    ], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    with open('tests/testdata/tweets.json', 'rb') as f:
        (stdout, stderr) = p.communicate(f.read())
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


@mark.posix
def test_tweets2concrete_stdout(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json',
        '-'
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    with open(output_file, 'wb') as f:
        f.write(stdout)

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_multiproc(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--num-proc', '2',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_log_every(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--log-level', 'INFO',
        '--log-interval', '1',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0
    assert len([line
                for line in stderr.decode('utf-8').strip().split('\n')
                if 'INFO' in line]) >= 2

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_unicode(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.unicode.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)
    assert validate_communication(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_gz(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_incomplete_gz(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--catch-ioerror',
        'tests/testdata/tweets.json.incomplete.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_incomplete_gz_multiproc(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--num-proc', '2',
        '--catch-ioerror',
        'tests/testdata/tweets.json.incomplete.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_log_config(log_conf, output_file):
    (log_conf_path, log_path) = log_conf
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--log-conf-path', log_conf_path,
        '--log-interval', '1',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert stdout.decode('utf-8') == ''
    assert stderr.decode('utf-8') == ''
    assert p.returncode == 0

    with open(log_path) as f:
        data = f.read()
        assert len([line
                    for line in data.strip().split('\n')
                    if 'INFO' in line]) >= 2

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_deleted(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.deleted.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_bad_line(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--skip-bad-lines',
        'tests/testdata/tweets.bad-line.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_bad_line_unicode(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--skip-bad-lines',
        'tests/testdata/tweets.bad-line-unicode.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_invalid(output_file):
    p = Popen([
        sys.executable,
        'scripts/tweets2concrete.py',
        '--skip-invalid-comms',
        'tests/testdata/tweets.invalid.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = next(it)
    assert_first_comm(comm)

    (comm, _) = next(it)
    assert_second_comm(comm)

    try:
        next(it)
    except StopIteration:
        pass
    else:
        assert False



from __future__ import unicode_literals
from pytest import raises, fixture

from math import exp, log

from concrete import (
    TokenizationKind, TokenLattice, LatticePath, Token,
    TokenTagging, TaggedToken, Tokenization, Arc,
    AnnotationMetadata
)

from concrete.util import create_comm
from concrete.util import (
    get_tokens, get_ner, get_pos, get_lemmas, get_tagged_tokens,
    compute_lattice_expected_counts, get_token_taggings
)

import mock


def allclose(x, y, rel_tol=1e-6, abs_tol=1e-9):
    return (
        len(x) == len(y) and
        len(x) == sum(
            abs(x[i] - y[i]) <= rel_tol * abs(x[i]) + abs_tol
            for i in range(len(x))
        )
    )


@fixture
def tokenization(request):
    return Tokenization(
        tokenTaggingList=[
            TokenTagging(
                metadata=AnnotationMetadata(tool='x'),
                taggingType='?',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='?'),
                    TaggedToken(tokenIndex=1, tag='?'),
                    TaggedToken(tokenIndex=2, tag='?'),
                ],
            ),
            TokenTagging(
                metadata=AnnotationMetadata(tool='x'),
                taggingType='POS',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='N'),
                    TaggedToken(tokenIndex=1, tag='N'),
                    TaggedToken(tokenIndex=2, tag='X'),
                ],
            ),
            TokenTagging(
                metadata=AnnotationMetadata(tool='y'),
                taggingType='NUMERAL',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='N'),
                    TaggedToken(tokenIndex=1, tag='N'),
                    TaggedToken(tokenIndex=2, tag='Y'),
                ],
            ),
            TokenTagging(
                metadata=AnnotationMetadata(tool='y'),
                taggingType='LEMMA',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='mambo'),
                    TaggedToken(tokenIndex=1, tag='number'),
                    TaggedToken(tokenIndex=2, tag='4'),
                ],
            ),
        ],
    )


def test_get_tokens_invalid_kind():
    with raises(ValueError):
        get_tokens(Tokenization(kind='invalid-kind'))


@mock.patch('concrete.util.tokenization.get_tagged_tokens')
def test_get_pos(mock_get_tagged_tokens):
    tokenization = mock.sentinel
    tool = mock.sentinel
    get_pos(tokenization, tool=tool)
    mock_get_tagged_tokens.assert_called_with(tokenization, 'POS', tool=tool)


@mock.patch('concrete.util.tokenization.get_tagged_tokens')
def test_get_lemmas(mock_get_tagged_tokens):
    tokenization = mock.sentinel
    tool = mock.sentinel
    get_lemmas(tokenization, tool=tool)
    mock_get_tagged_tokens.assert_called_with(tokenization, 'LEMMA', tool=tool)


@mock.patch('concrete.util.tokenization.get_tagged_tokens')
def test_get_ner(mock_get_tagged_tokens):
    tokenization = mock.sentinel
    tool = mock.sentinel
    get_ner(tokenization, tool=tool)
    mock_get_tagged_tokens.assert_called_with(tokenization, 'NER', tool=tool)


def test_get_tagged_tokens(tokenization):
    assert ['N', 'N', 'Y'] == list(map(
        lambda t: t.tag,
        get_tagged_tokens(tokenization, 'NUMERAL')))
    assert [0, 1, 2] == list(map(
        lambda t: t.tokenIndex,
        get_tagged_tokens(tokenization, 'NUMERAL')))


def test_get_tagged_tokens_lowercase(tokenization):
    assert ['N', 'N', 'Y'] == list(map(
        lambda t: t.tag,
        get_tagged_tokens(tokenization, 'numeral')))
    assert [0, 1, 2] == list(map(
        lambda t: t.tokenIndex,
        get_tagged_tokens(tokenization, 'numeral')))


def test_get_tagged_tokens_no_tagging(tokenization):
    tokenization.tokenTaggingList = filter(
        lambda ttl: ttl.taggingType != 'NUMERAL',
        tokenization.tokenTaggingList
    )
    with raises(Exception):
        get_tagged_tokens(tokenization, 'NUMERAL')


def test_get_tagged_tokens_non_unique_tagging(tokenization):
    tokenization.tokenTaggingList.append(
        TokenTagging(
            taggingType='NUMERAL',
            taggedTokenList=[
                TaggedToken(tokenIndex=0, tag='N'),
                TaggedToken(tokenIndex=1, tag='Y'),
                TaggedToken(tokenIndex=2, tag='Y'),
            ],
        ),
    )
    with raises(Exception):
        get_tagged_tokens(tokenization, 'NUMERAL')


def test_get_tagged_tokens_non_unique_tagging_specify_tool(tokenization):
    tokenization.tokenTaggingList.append(
        TokenTagging(
            metadata=AnnotationMetadata(tool='z'),
            taggingType='NUMERAL',
            taggedTokenList=[
                TaggedToken(tokenIndex=0, tag='N'),
                TaggedToken(tokenIndex=1, tag='Y'),
                TaggedToken(tokenIndex=2, tag='Y'),
            ],
        ),
    )
    assert ['N', 'N', 'Y'] == list(map(
        lambda t: t.tag,
        get_tagged_tokens(tokenization, 'NUMERAL', tool='y')))
    assert [0, 1, 2] == list(map(
        lambda t: t.tokenIndex,
        get_tagged_tokens(tokenization, 'NUMERAL', tool='y')))


def test_get_tagged_tokens_non_unique_tagging_specify_tool_uppercase(tokenization):
    tokenization.tokenTaggingList.append(
        TokenTagging(
            metadata=AnnotationMetadata(tool='Z'),
            taggingType='NUMERAL',
            taggedTokenList=[
                TaggedToken(tokenIndex=0, tag='N'),
                TaggedToken(tokenIndex=1, tag='Y'),
                TaggedToken(tokenIndex=2, tag='Y'),
            ],
        ),
    )
    assert ['N', 'N', 'Y'] == list(map(
        lambda t: t.tag,
        get_tagged_tokens(tokenization, 'NUMERAL', tool='y')))
    assert [0, 1, 2] == list(map(
        lambda t: t.tokenIndex,
        get_tagged_tokens(tokenization, 'NUMERAL', tool='y')))


def test_get_tagged_tokens_no_tagging_specify_tool(tokenization):
    with raises(Exception):
        get_tagged_tokens(tokenization, 'NUMERAL', tool='z')


def test_get_token_taggings(tokenization):
    assert [[(0, 'N'), (1, 'N'), (2, 'Y')]] == [
        [(t.tokenIndex, t.tag) for t in tt.taggedTokenList]
        for tt in get_token_taggings(tokenization, 'NUMERAL')
    ]


def test_get_token_taggings_lowercase(tokenization):
    assert [[(0, 'N'), (1, 'N'), (2, 'Y')]] == [
        [(t.tokenIndex, t.tag) for t in tt.taggedTokenList]
        for tt in get_token_taggings(tokenization, 'numeral')
    ]


def test_get_token_taggings_lowercase_case_sensitive(tokenization):
    assert [] == get_token_taggings(tokenization, 'numeral',
                                    case_sensitive=True)


def test_get_token_taggings_no_tagging(tokenization):
    assert [] == get_token_taggings(tokenization, '!NUMERAL')


def test_get_token_taggings_non_unique_tagging(tokenization):
    tokenization.tokenTaggingList.append(
        TokenTagging(
            taggingType='NUMERAL',
            taggedTokenList=[
                TaggedToken(tokenIndex=0, tag='N'),
                TaggedToken(tokenIndex=1, tag='Y'),
                TaggedToken(tokenIndex=2, tag='Y'),
            ],
        ),
    )
    assert [
        [(0, 'N'), (1, 'N'), (2, 'Y')],
        [(0, 'N'), (1, 'Y'), (2, 'Y')],
    ] == [
        [(t.tokenIndex, t.tag) for t in tt.taggedTokenList]
        for tt in get_token_taggings(tokenization, 'NUMERAL')
    ]


def test_no_lattice():
    comm = create_comm('comm-1', 'mambo no. 4')
    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    tokenization.kind = None
    token_texts = [t.text for t in get_tokens(tokenization)]
    assert ['mambo', 'no.', '4'] == token_texts


def test_no_lattice_with_no_kind():
    comm = create_comm('comm-1', 'mambo no. 4')
    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    token_texts = [t.text for t in get_tokens(tokenization)]
    assert ['mambo', 'no.', '4'] == token_texts


def test_lattice_with_token_list_kind():
    comm = create_comm('comm-1', 'mambo no. 4')
    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    lattice_path = LatticePath()
    lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                              Token(tokenIndex=0, text='no.'),
                              Token(tokenIndex=0, text='3')]
    token_lattice = TokenLattice()
    token_lattice.cachedBestPath = lattice_path
    tokenization.lattice = token_lattice
    token_texts = [t.text for t in get_tokens(tokenization)]
    assert ['mambo', 'no.', '4'] == token_texts


def test_lattice_with_token_lattice_kind():
    comm = create_comm('comm-1', 'mambo no. 4')
    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    lattice_path = LatticePath()
    lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                              Token(tokenIndex=0, text='no.'),
                              Token(tokenIndex=0, text='3')]
    token_lattice = TokenLattice()
    token_lattice.cachedBestPath = lattice_path
    tokenization.lattice = token_lattice
    tokenization.kind = TokenizationKind.TOKEN_LATTICE
    token_texts = [t.text for t in get_tokens(tokenization)]
    assert ['mambo', 'no.', '3'] == token_texts


def test_lattice_with_no_kind():
    comm = create_comm('comm-1', 'mambo no. 4')
    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    lattice_path = LatticePath()
    lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                              Token(tokenIndex=0, text='no.'),
                              Token(tokenIndex=0, text='3')]
    token_lattice = TokenLattice()
    token_lattice.cachedBestPath = lattice_path
    tokenization.lattice = token_lattice
    tokenization.kind = None
    token_texts = [t.text for t in get_tokens(tokenization)]
    assert ['mambo', 'no.', '4'] == token_texts


def test_compute_lattice_expected_counts_empty():
    assert compute_lattice_expected_counts(TokenLattice(arcList=[
    ], startState=0, endState=0)) == []


def test_compute_lattice_expected_counts_one_arc():
    # 0 --0-> 1, -1
    expected = [0.]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-1.),
    ], startState=0, endState=1))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_two_parallel_arcs():
    # 0 --0-> 1, -1
    # 0 --0-> 1, -3
    expected = [0.]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-1.),
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-3.),
    ], startState=0, endState=1))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_two_serial_arcs():
    # 0 --0-> 1, -1
    # 1 --1-> 2, -3
    expected = [0., 0.]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-1.),
        Arc(src=1, dst=2, token=Token(tokenIndex=1), weight=-3.),
    ], startState=0, endState=2))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_triangle():
    # 0 --0-> 1, -1
    # 1 --1-> 2, -2
    # 0 --1-> 2, -4
    A = log(exp(-1 + -2) + exp(-4))
    expected = [-1 + -2 - A, 0.]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-1.),
        Arc(src=1, dst=2, token=Token(tokenIndex=1), weight=-2.),
        Arc(src=0, dst=2, token=Token(tokenIndex=1), weight=-4.),
    ], startState=0, endState=2))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_triangle_arbitrary_states():
    # 47 --0-> 9, -1
    #  9 --1-> 3, -2
    # 47 --1-> 3, -4
    A = log(exp(-1 + -2) + exp(-4))
    expected = [-1 + -2 - A, 0.]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=47, dst=9, token=Token(tokenIndex=0), weight=-1.),
        Arc(src=9, dst=3, token=Token(tokenIndex=1), weight=-2.),
        Arc(src=47, dst=3, token=Token(tokenIndex=1), weight=-4.),
    ], startState=47, endState=3))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_rhombus():
    # 0 --0-> 1, -1
    # 1 --1-> 3, -2
    # 0 --0-> 2, -3
    # 2 --2-> 3, -4
    A = log(exp(-1 + -2) + exp(-3 + -4))
    expected = [0., -1 + -2 - A, -3 + -4 - A]
    actual = compute_lattice_expected_counts(TokenLattice(arcList=[
        Arc(src=0, dst=1, token=Token(tokenIndex=0), weight=-1.),
        Arc(src=1, dst=3, token=Token(tokenIndex=1), weight=-2.),
        Arc(src=0, dst=2, token=Token(tokenIndex=0), weight=-3.),
        Arc(src=2, dst=3, token=Token(tokenIndex=2), weight=-4.),
    ], startState=0, endState=3))
    assert allclose(expected, actual), '%s !~= %s' % (expected, actual)


def test_compute_lattice_expected_counts_incomplete_arc():
    # 0 --0-> 1, -1
    with raises(ValueError):
        compute_lattice_expected_counts(TokenLattice(arcList=[
            Arc(dst=1, token=Token(tokenIndex=0), weight=-1.),
        ], startState=0, endState=1))
    with raises(ValueError):
        compute_lattice_expected_counts(TokenLattice(arcList=[
            Arc(src=0, token=Token(tokenIndex=0), weight=-1.),
        ], startState=0, endState=1))
    with raises(ValueError):
        compute_lattice_expected_counts(TokenLattice(arcList=[
            Arc(src=0, dst=1, weight=-1.),
        ], startState=0, endState=1))
    with raises(ValueError):
        compute_lattice_expected_counts(TokenLattice(arcList=[
            Arc(src=0, dst=1, token=Token(tokenIndex=0)),
        ], startState=0, endState=1))



from __future__ import unicode_literals
import time

from testfixtures import LogCapture, StringComparison

import concrete
from concrete.util import add_references_to_communication
from concrete.validate import (
    validate_communication, validate_entity_mention_ids,
    validate_entity_mention_tokenization_ids,
    validate_token_offsets_for_section,
    validate_thrift_deep,
    validate_token_offsets_for_sentence
)
from concrete import (
    Section,
    Sentence,
    TextSpan,
    Token,
    TokenList,
    Tokenization,
)
from test_helper import read_test_comm


def create_section_with_sentence(section_start, section_ending,
                                 sentence_start, sentence_ending):
    sentence_textspan = TextSpan(
        start=sentence_start, ending=sentence_ending)
    sentence = Sentence(
        textSpan=sentence_textspan, uuid='TEST_SENTENCE')
    section_textspan = TextSpan(
        start=section_start, ending=section_ending)
    section = Section(
        sentenceList=[sentence],
        textSpan=section_textspan,
        uuid='TEST_SECTION')
    return section


def create_sentence_with_token(sentence_start, sentence_ending,
                               token_start, token_ending):
    token_textspan = TextSpan(
        start=token_start, ending=token_ending)
    token = Token(textSpan=token_textspan)
    tokenization = Tokenization(
        tokenList=TokenList(tokenList=[token]))
    sentence_textspan = TextSpan(
        start=sentence_start, ending=sentence_ending)
    sentence = Sentence(
        tokenization=tokenization,
        textSpan=sentence_textspan,
        uuid='TEST')
    return sentence


def test_add_references():
    comm = read_test_comm()
    add_references_to_communication(comm)


def test_entity_mention_ids():
    comm = read_test_comm()
    assert validate_communication(comm)
    assert validate_entity_mention_ids(comm)

    comm.entitySetList[0].entityList[0].mentionIdList[
        0] = concrete.UUID(uuidString='BAD_ENTITY_MENTION_UUID')

    with LogCapture() as log_capture:
        assert not validate_entity_mention_ids(comm)
    log_capture.check(('root', 'ERROR', StringComparison(
        r'.*invalid entityMentionId.*BAD_ENTITY_MENTION_UUID')))


def test_entity_mention_tokenization():
    comm = read_test_comm()
    assert validate_communication(comm)
    assert validate_entity_mention_ids(comm)

    comm.entityMentionSetList[0].mentionList[0].tokens.tokenizationId = (
        concrete.UUID(uuidString='BAD_TOKENIZATION_UUID')
    )

    with LogCapture() as log_capture:
        assert not validate_entity_mention_tokenization_ids(comm)
    log_capture.check(('root', 'ERROR', StringComparison(
        r'.*invalid tokenizationId.*BAD_TOKENIZATION_UUID')))


def test_check_required_fields():
    # When a field is marked as required in a .thrift file, the
    # Python code generated by the Thrift compiler only seems to
    # capture this requirement in the validate() function for the
    # generated class.  While the ThriftGeneratedClass.thrift_spec
    # structure captures the names and types of the fields,
    # thrift_spec does not seem to store any flags indicating
    # whether or not a field is required.
    #
    # Here is the validate() function for the Communication class:
    #
    #    def validate():
    #        if self.id is None:
    #            raise TProtocol.TProtocolException(
    #                message='Required field id is unset!')
    #        if self.uuid is None:
    #            raise TProtocol.TProtocolException(
    #                message='Required field uuid is unset!')
    #        if self.type is None:
    #            raise TProtocol.TProtocolException(
    #                message='Required field type is unset!')
    #        return
    #
    # The validate() function raises an exception when it can't
    # find a required field.  There doesn't seem to be any way to
    # determine whether multiple required fields are missing,
    # aside from assigning a value to the required field and
    # running validate() again.

    comm = concrete.Communication()

    with LogCapture() as log_capture:
        assert not validate_thrift_deep(comm)
    log_capture.check(
        ('root', 'ERROR',
         "Communication: Required Field 'id' is unset!"))

    comm.id = "ID"
    with LogCapture() as log_capture:
        assert not validate_thrift_deep(comm)
    log_capture.check(
        ('root', 'ERROR',
         "Communication: Required Field 'uuid' is unset!"))

    comm.uuid = concrete.UUID(uuidString="TEST_UUID")
    with LogCapture() as log_capture:
        assert not validate_thrift_deep(comm)
    log_capture.check(('root', 'ERROR', StringComparison(
        r".*TEST_UUID.*Required Field 'type' is unset!")))

    comm.metadata = concrete.AnnotationMetadata(
        tool="TEST", timestamp=int(time.time()))

    comm.type = "OTHER"
    assert validate_thrift_deep(comm)


def test_validate_token_offsets_for_good_sentence():
    sentence = create_sentence_with_token(0, 30, 0, 10)
    assert validate_token_offsets_for_sentence(sentence)


def test_validate_token_offsets_for_sentence_rev_sentence_offsets():
    sentence = create_sentence_with_token(30, 0, 0, 10)
    with LogCapture():
        assert not validate_token_offsets_for_sentence(sentence)


def test_validate_token_offsets_for_sentence_rev_token_offsets():
    sentence = create_sentence_with_token(0, 30, 10, 0)
    with LogCapture():
        assert not validate_token_offsets_for_sentence(sentence)


def test_validate_token_offsets_for_sentence_token_outside():
    sentence = create_sentence_with_token(0, 30, 25, 35)
    with LogCapture():
        assert not validate_token_offsets_for_sentence(sentence)


def test_validate_token_offsets_for_good_section():
    section = create_section_with_sentence(0, 30, 0, 10)
    assert validate_token_offsets_for_section(section)


def test_validate_token_offsets_for_section_rev_section_offsets():
    section = create_section_with_sentence(30, 0, 0, 10)
    with LogCapture():
        assert not validate_token_offsets_for_section(section)


def test_validate_token_offsets_for_section_rev_token_offsets():
    section = create_section_with_sentence(0, 30, 10, 0)
    with LogCapture():
        assert not validate_token_offsets_for_section(section)


def test_validate_token_offsets_for_section_sentence_outside():
    section = create_section_with_sentence(0, 30, 25, 35)
    with LogCapture():
        assert not validate_token_offsets_for_section(section)


def test_validate_token_offsets_for_real_example_section_data():
    section = create_section_with_sentence(55, 296, 0, 118)
    with LogCapture():
        assert not validate_token_offsets_for_section(section)



from __future__ import unicode_literals
import logging
import os
from hashlib import md5

from ..access.ttypes import FetchResult
from ..services.ttypes import ServiceInfo
from .access_wrapper import FetchCommunicationClientWrapper
from .file_io import write_communication_to_file
from .mem_io import write_communication_to_buffer
from .redis_io import RedisCommunicationWriter
from ..version import concrete_library_version


DEFAULT_S3_KEY_PREFIX_LEN = 4


class CommunicationContainerFetchHandler(object):
    """FetchCommunicationService implementation using Communication containers

    Implements the :mod:`.FetchCommunicationService` interface, retrieving
    Communications from a dict-like `communication_container` object
    that maps Communication ID strings to Communications.  The
    `communication_container` could be an actual dict, or a container
    such as:

    - :class:`.DirectoryBackedCommunicationContainer`
    - :class:`.FetchBackedCommunicationContainer`
    - :class:`.MemoryBackedCommunicationContainer`
    - :class:`.RedisHashBackedCommunicationContainer`
    - :class:`.ZipFileBackedCommunicationContainer`
    - :class:`.S3BackedCommunicationContainer`

    Usage::

        from concrete.util.access_wrapper import FetchCommunicationServiceWrapper

        handler = CommunicationContainerFetchHandler(comm_container)
        fetch_service = FetchCommunicationServiceWrapper(handler)
        fetch_service.serve(host, port)
    """

    def __init__(self, communication_container):
        """
        Args:
            communication_container: Dict-like object that maps Communication
                                     IDs to Communications
        """
        self.communication_container = communication_container

    def about(self):
        logging.debug("Received about() call")
        service_info = ServiceInfo()
        service_info.name = 'CommunicationContainerFetchHandler - %s' % \
                            type(self.communication_container)
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        logging.debug("Received alive() call")
        return True

    def fetch(self, fetch_request):
        logging.debug("Received FetchRequest: %s" % fetch_request)
        fetch_result = FetchResult()
        fetch_result.communications = []
        for communication_id in fetch_request.communicationIds:
            if communication_id in self.communication_container:
                comm = self.communication_container[communication_id]
                fetch_result.communications.append(comm)
            else:
                logging.warning('Unable to find Communication with ID: %s' % communication_id)
        return fetch_result

    def getCommunicationCount(self):
        logging.debug('Received getCommunicationCount()')
        communicationCount = len(self.communication_container)
        logging.debug('- Communication Count: %d' % communicationCount)
        return communicationCount

    def getCommunicationIDs(self, offset, count):
        logging.debug('Received getCommunicationIDs() call')
        return list(self.communication_container.keys())[offset:][:count]


class DirectoryBackedStoreHandler(object):
    """Simple StoreCommunicationService implementation using a directory

    Implements the :mod:`.StoreCommunicationService` interface, storing
    Communications in a directory.
    """
    def __init__(self, store_path):
        """
        Args:
            store_path: Path where Communications should be Stored
        """
        self.store_path = store_path

    def about(self):
        logging.debug("DirectoryBackedStoreHandler.about() called")
        service_info = ServiceInfo()
        service_info.name = 'DirectoryBackedStoreHandler'
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        logging.debug("DirectoryBackedStoreHandler.alive() called")
        return True

    def store(self, communication):
        """Save Communication to a directory

        Stored Communication files will be named `[COMMUNICATION_ID].comm`.
        If a file with that name already exists, it will be overwritten.
        """
        logging.debug(
            "DirectoryBackedStoreHandler.store() called with Communication "
            "with ID '%s'" % communication.id)
        comm_filename = os.path.join(self.store_path, communication.id + '.comm')
        write_communication_to_file(communication, comm_filename)


class RelayFetchHandler(object):
    """Implements a 'relay' to another :mod:`.FetchCommunicationService`
    server.

    A :mod:`.FetchCommunicationService` that acts as a relay to a
    second :mod:`.FetchCommunicationService`, where the second
    service is using the TSocket transport and TCompactProtocol
    protocol.

    This class was designed for the use case where you have Thrift
    JavaScript code that needs to communicate with a
    :mod:`.FetchCommunicationService` server, but the server does
    not support the same Thrift serialization protocol as the
    JavaScript client.

    The de-facto standard for Concrete services is to use the
    TCompactProtocol serialization protocol over a TSocket connection.
    But as of Thrift 0.10.0, the Thrift JavaScript libraries only
    support using TJSONProtocol over HTTP.

    The RelayFetchHandler class is intended to be used as server-side
    code by a web application.  The JavaScript code will make
    :mod:`.FetchCommunicationService` RPC calls to the web server
    using HTTP/TJSONProtocol, and the web application will then pass
    these RPC calls to another :mod:`.FetchCommunicationService`
    using TSocket/TCompactProtocol RPC calls.
    """
    def __init__(self, host, port):
        """
        Args:
            host (str): Hostname of :mod:`.FetchCommunicationService` server
            port (int): Port # of :mod:`.FetchCommunicationService` server
        """
        self.host = host
        self.port = int(port)

    def about(self):
        logging.debug('RelayFetchHandler.about()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.about()

    def alive(self):
        logging.debug('RelayFetchHandler.alive()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.alive()

    def fetch(self, request):
        logging.debug('RelayFetchHandler.fetch()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.fetch(request)

    def getCommunicationCount(self):
        logging.debug('RelayFetchHandler.getCommunicationCount()')
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationCount()

    def getCommunicationIDs(self, offset, count):
        logging.debug('RelayFetchHandler.getCommunicationIDs(offset=%d, count=%d)' %
                      (offset, count))
        with FetchCommunicationClientWrapper(self.host, self.port) as fc:
            return fc.getCommunicationIDs(offset, count)


def prefix_s3_key(key_str, prefix_len):
    '''
    Given unprefixed S3 key `key_str`, prefix the key with a
    deterministic prefix of hex characters of length `prefix_len` and
    return the result.  Keys with such prefixes enable better
    performance on S3 and reduce the likelihood of rate-limiting.

    Args:
        key_str (str): original (unprefixed) key, as a string
        prefix_len (int): length of prefix to add to key

    Returns:
        prefixed key

    References:
        http://docs.aws.amazon.com/AmazonS3/latest/dev/request-rate-perf-considerations.html
    '''
    return md5(key_str.encode('utf-8')).hexdigest()[:prefix_len] + key_str


def unprefix_s3_key(prefixed_key_str, prefix_len):
    '''
    Given prefixed S3 key `key_str`, remove prefix of length
    `prefix_len` from the key and return the result.
    Keys with random-looking prefixes enable better performance on S3
    and reduce the likelihood of rate-limiting.

    Args:
        preixed_key_str (str): prefixed key, as a string
        prefix_len (int): length of prefix to remove from key

    Returns:
        unprefixed key

    References:
        http://docs.aws.amazon.com/AmazonS3/latest/dev/request-rate-perf-considerations.html
    '''
    return prefixed_key_str[prefix_len:]


class S3BackedStoreHandler(object):
    """Simple StoreCommunicationService implementation using an AWS S3
    bucket.

    Implements the :mod:`.StoreCommunicationService` interface, storing
    Communications in an S3 bucket, indexed by id, optionally prefixed
    with a fixed-length, random-looking but deterministic hash to
    improve performance.

    References:
        http://docs.aws.amazon.com/AmazonS3/latest/dev/request-rate-perf-considerations.html
    """
    def __init__(self, bucket, prefix_len=DEFAULT_S3_KEY_PREFIX_LEN):
        """
        Args:
            bucket (boto.s3.bucket.Bucket): S3 bucket object
            prefix_len (int): length of prefix to add to
                a Communication id to form its key.  A
                prefix of length four enables S3 to better partition the
                bucket contents, yielding higher performance and a lower
                chance of getting rate-limited by AWS.
        """
        self.bucket = bucket
        self.prefix_len = prefix_len

    def about(self):
        """
        Return S3BackedStoreHandler service information.

        Returns:
            An object of type :class:`.ServiceInfo`
        """
        logging.debug("S3BackedStoreHandler.about() called")
        service_info = ServiceInfo()
        service_info.name = 'S3BackedStoreHandler'
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        """
        Return whether service is alive and running.

        Returns:
            True or False
        """
        logging.debug("S3BackedStoreHandler.alive() called")
        return True

    def store(self, communication):
        """
        Save Communication to an S3 bucket, using the Communication
        id with a hash prefix of length `self.prefix_len`
        as a key.

        Args:
            communication (Communication): communication to store
        """
        logging.debug(
            "S3BackedStoreHandler.store() called with Communication "
            "with ID '%s'" % communication.id)
        buf = write_communication_to_buffer(communication)
        prefixed_key_str = prefix_s3_key(communication.id, self.prefix_len)
        key = self.bucket.get_key(prefixed_key_str, validate=False)
        key.set_contents_from_string(buf)


class RedisHashBackedStoreHandler(object):
    """Simple StoreCommunicationService implementation using a Redis
    hash.

    Implements the :mod:`.StoreCommunicationService` interface, storing
    Communications in a Redis hash, indexed by id.
    """
    def __init__(self, redis_db, key):
        """
        Args:
            redis_db (redis.Redis): Redis database connection object
            key (str): key of hash in redis database
        """
        self.writer = RedisCommunicationWriter(redis_db, key, key_type='hash')

    def about(self):
        logging.debug("RedisHashBackedStoreHandler.about() called")
        service_info = ServiceInfo()
        service_info.name = 'RedisHashBackedStoreHandler'
        service_info.version = concrete_library_version()
        return service_info

    def alive(self):
        logging.debug("RedisHashBackedStoreHandler.alive() called")
        return True

    def store(self, communication):
        """Save Communication to a Redis hash, using the Communication
        id as a key.

        Args:
            communication (Communication): communication to store
        """
        logging.debug(
            "RedisHashBackedStoreHandler.store() called with Communication "
            "with ID '%s'" % communication.id)
        self.writer.write(communication)



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class Context(object):
    """
    A structure intended to convey context about a particular
    Concrete communication.

    Contexts are intended to be used to convey meta-communication
    information to analytics via an RPC method. It is expected that
    services consuming or producing Contexts are coupled,
    delivering an agreed upon format that is capable of
    being interpreted and processed between two particular services.

    Currently, it is being used to transmit hypotheses alongside
    concrete communications for AIDA.

    Attributes:
     - contents: The contents of the Context. Services should agree
    upon what the expected format of the contents are
    (e.g. JSON, RDF) between themselves.

    """


    def __init__(self, contents=None,):
        self.contents = contents

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.contents = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Context')
        if self.contents is not None:
            oprot.writeFieldBegin('contents', TType.STRING, 1)
            oprot.writeString(self.contents.encode('utf-8') if sys.version_info[0] == 2 else self.contents)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Context)
Context.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'contents', 'UTF8', None, ),  # 1
)
fix_spec(all_structs)
del all_structs



from __future__ import unicode_literals
from thrift.transport import TTransport
from thrift.transport import TSocket
from thrift.protocol import TCompactProtocol

from concrete.util import (
    SubprocessAnnotateCommunicationServiceWrapper
)
from concrete.util import find_port
from concrete.util import create_comm


from time import time

from concrete.annotate import AnnotateCommunicationService
from concrete import AnnotationMetadata


class NoopAnnotateCommunicationService(AnnotateCommunicationService.Iface):
    METADATA_TOOL = 'No-op AnnotateCommunicationService'

    def annotate(self, communication):
        return communication

    def getMetadata(self,):
        metadata = AnnotationMetadata(tool=self.METADATA_TOOL,
                                      timestamp=int(time()))
        return metadata

    def getDocumentation(self):
        return '''\
        AnnotateCommunicationService that returns communication unmodified
        '''

    def shutdown(self):
        pass


def test_annotate():
    impl = NoopAnnotateCommunicationService()
    host = '127.0.0.1'
    port = find_port()
    timeout = 5

    comm_id = '1-2-3-4'
    comm = create_comm(comm_id)

    comm_uuid_uuidString = comm.uuid.uuidString
    comm_metadata_tool = comm.metadata.tool
    comm_metadata_timestamp = comm.metadata.timestamp

    with SubprocessAnnotateCommunicationServiceWrapper(impl, host, port,
                                                       timeout=timeout):
        transport = TSocket.TSocket(host, port)
        transport = TTransport.TFramedTransport(transport)
        protocol = TCompactProtocol.TCompactProtocolAccelerated(transport)

        cli = AnnotateCommunicationService.Client(protocol)
        transport.open()
        res = cli.annotate(comm)
        transport.close()

        assert res.id == comm_id
        assert res.uuid.uuidString == comm_uuid_uuidString
        assert res.metadata.tool == comm_metadata_tool
        assert res.metadata.timestamp == comm_metadata_timestamp


def test_get_metadata():
    impl = NoopAnnotateCommunicationService()
    host = '127.0.0.1'
    port = find_port()
    timeout = 5

    with SubprocessAnnotateCommunicationServiceWrapper(impl, host, port,
                                                       timeout=timeout):
        transport = TSocket.TSocket(host, port)
        transport = TTransport.TFramedTransport(transport)
        protocol = TCompactProtocol.TCompactProtocolAccelerated(transport)

        cli = AnnotateCommunicationService.Client(protocol)
        transport.open()
        metadata = cli.getMetadata()
        transport.close()

        assert NoopAnnotateCommunicationService.METADATA_TOOL == metadata.tool



#!/usr/bin/env python

import sys

from thrift.transport import TTransport
from thrift.protocol import TCompactProtocol

try:
    transport = TTransport.TMemoryBuffer()
    TCompactProtocol.TCompactProtocolAccelerated(transport, fallback=False)
    sys.stderr.write('Thrift is accelerated\n')
    sys.exit(0)
except Exception:
    sys.stderr.write('ERROR: Thrift is NOT accelerated\n')
    sys.exit(-1)



#!/usr/bin/env python


from __future__ import print_function
from __future__ import unicode_literals
import json
import time

import concrete
import concrete.util
from concrete.validate import validate_communication
from concrete.util.concrete_uuid import AnalyticUUIDGeneratorFactory


def main():
    comm = create_comm_from_tweet(JSON_TWEET_STRING)
    concrete.util.write_communication_to_file(comm, "tweet.comm")
    add_dictionary_tagging(comm)
    concrete.util.write_communication_to_file(comm, "tweet_pos.comm")


def create_comm_from_tweet(json_tweet_string):
    """Create a Concrete Communication from a JSON Tweet string

    Args:
        json_tweet_string: A JSON string for a Tweet, using the JSON
            format specified by the Twitter API:
              https://dev.twitter.com/docs/platform-objects/tweets

    Returns:
        A Concrete Communication object
    """
    tweet_data = json.loads(json_tweet_string)

    augf = AnalyticUUIDGeneratorFactory()
    aug = augf.create()

    comm = concrete.Communication()
    comm.id = "Annotation_Test_1"
    comm.metadata = concrete.AnnotationMetadata(
        tool="Annotation Example script",
        timestamp=int(time.time())
    )
    comm.text = tweet_data['text']
    comm.type = "Tweet"
    comm.uuid = next(aug)

    comm.sectionList = [concrete.Section()]
    comm.sectionList[0].kind = "mySectionKind"
    comm.sectionList[0].uuid = next(aug)
    comm.sectionList[0].sentenceList = [concrete.Sentence()]
    comm.sectionList[0].sentenceList[0].uuid = next(aug)
    comm.sectionList[0].sentenceList[0].tokenization = concrete.Tokenization()

    tokenization = comm.sectionList[0].sentenceList[0].tokenization
    tokenization.kind = concrete.TokenizationKind.TOKEN_LIST
    tokenization.metadata = concrete.AnnotationMetadata(
        tool="TEST", timestamp=int(time.time()))
    tokenization.tokenList = concrete.TokenList()
    tokenization.tokenList.tokenList = []
    tokenization.uuid = next(aug)

    # Whitespace tokenization
    tokens = comm.text.split()

    for i, token_text in enumerate(tokens):
        t = concrete.Token()
        t.tokenIndex = i
        t.text = token_text
        tokenization.tokenList.tokenList.append(t)

    if validate_communication(comm):
        print("Created valid Communication")
    else:
        print("ERROR: Invalid Communication")

    return comm


def add_dictionary_tagging(comm):
    """Adds In/Out of dictionary 'POS' tags to a Communication

    Takes a Concrete Communication, adds a Part-Of-Speech tag to each
    token, where the tags record whether the token is 'In' or 'Out' of
    the system dictionary.

    Args:
        comm: A Concrete Communication with tokens

    Returns:
        A copy of the original Communication, with POS tags added
    """
    dictionary = set()
    for w in open('/usr/share/dict/words'):
        dictionary.add(w.strip().lower())

    augf = AnalyticUUIDGeneratorFactory(comm)
    aug = augf.create()

    if comm.sectionList:
        for section in comm.sectionList:
            if section.sentenceList:
                for sentence in section.sentenceList:
                    posTagList = concrete.TokenTagging()
                    posTagList.metadata = concrete.AnnotationMetadata(
                        tool="POS Tagger", timestamp=int(time.time()))
                    posTagList.taggingType = "POS"
                    posTagList.taggedTokenList = []
                    posTagList.uuid = next(aug)
                    tkzn = sentence.tokenization
                    if tkzn.tokenList:
                        for i, token in enumerate(tkzn.tokenList.tokenList):
                            tt = concrete.TaggedToken()
                            tt.tokenIndex = i
                            if token.text.lower() in dictionary:
                                tt.tag = "In"
                            else:
                                tt.tag = "Out"
                            posTagList.taggedTokenList.append(tt)
                            print("%d [%s] %s" % (i, token.text, tt.tag))
                    tkzn.tokenTaggingList = [posTagList]
            print()

    if validate_communication(comm):
        print("Created valid POS tagging for Communication")
    else:
        print("ERROR: Invalid POS tagging Communication")
    return comm


JSON_TWEET_STRING = '''{"contributors": null, "truncated": false, "text": "Bar\
ber tells me - his son is colorblind / my hair is auburn / and auburn is a sha\
de of green", "in_reply_to_status_id": null, "id": 238426131689242624, "favori\
te_count": 0, "source": "<a href=\\"http://twitter.com\\" rel=\\"nofollow\\">T\
witter Web Client</a>", "retweeted": false, "coordinates": null, "entities": {\
"symbols": [], "user_mentions": [], "hashtags": [], "urls": []}, "in_reply_to_\
screen_name": null, "id_str": "238426131689242624", "retweet_count": 0, "in_re\
ply_to_user_id": null, "favorited": false, "user": {"follow_request_sent": nul\
l, "profile_use_background_image": true, "default_profile_image": false, "id":\
 18063351, "profile_background_image_url_https": "https://abs.twimg.com/images\
/themes/theme5/bg.gif", "verified": false, "profile_text_color": "3E4415", "pr\
ofile_image_url_https": "https://pbs.twimg.com/profile_images/67158916/n370391\
7_32092098_7623_normal.jpg", "profile_sidebar_fill_color": "99CC33", "entities\
": {"url": {"urls": [{"url": "http://t.co/Qb6hKcbqgj", "indices": [0, 22], "ex\
panded_url": "http://craigharman.net", "display_url": "craigharman.net"}]}, "d\
escription": {"urls": []}}, "followers_count": 78, "profile_sidebar_border_col\
or": "829D5E", "id_str": "18063351", "profile_background_color": "352726", "li\
sted_count": 5, "is_translation_enabled": false, "utc_offset": -14400, "status\
es_count": 26, "description": "", "friends_count": 54, "location": "", "profil\
e_link_color": "D02B55", "profile_image_url": "http://pbs.twimg.com/profile_im\
ages/67158916/n3703917_32092098_7623_normal.jpg", "following": null, "geo_enab\
led": false, "profile_background_image_url": "http://abs.twimg.com/images/them\
es/theme5/bg.gif", "screen_name": "charman", "lang": "en", "profile_background\
_tile": false, "favourites_count": 0, "name": "Craig Harman", "notifications":\
 null, "url": "http://t.co/Qb6hKcbqgj", "created_at": "Thu Dec 11 23:07:27 +00\
00 2008", "contributors_enabled": false, "time_zone": "Eastern Time (US & Cana\
da)", "protected": false, "default_profile": false, "is_translator": false}, "\
geo": null, "in_reply_to_user_id_str": null, "lang": "en", "created_at": "Thu \
Aug 23 00:03:14 +0000 2012", "in_reply_to_status_id_str": null, "place": null}
'''


if __name__ == "__main__":
    main()



#!/usr/bin/env python


"""
Deprecated: please use validate-communication.py instead.

Command line script to (partially) validate a Concrete Communication

This script is a thin wrapper around the functionality in the
concrete.validate library.
"""
from __future__ import unicode_literals

import argparse
import logging
import sys

import concrete.version
from concrete.validate import validate_communication_file


sys.stderr.write(
    'validate_communication.py is deprecated '
    'and will be removed in the future.\n')
sys.stderr.write('Please use validate-communication.py instead.\n')


def main():
    parser = argparse.ArgumentParser(
        description="Validate a Concrete Communication file")
    parser.add_argument('communication_file')
    parser.add_argument('-l', '--loglevel', choices=['debug', 'info',
                                                     'warning', 'error'],
                        help="Log level threshold")
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    # Set logging level
    if args.loglevel:
        lowercase_loglevel = args.loglevel.lower()
        if lowercase_loglevel == 'debug':
            loglevel = logging.DEBUG
        elif lowercase_loglevel == 'info':
            loglevel = logging.INFO
        elif lowercase_loglevel == 'warning':
            loglevel = logging.WARNING
        elif lowercase_loglevel == 'error':
            loglevel = logging.ERROR
        else:
            loglevel = logging.DEBUG
    else:
        loglevel = logging.DEBUG
    logging.basicConfig(format='%(levelname)7s:  %(message)s', level=loglevel)

    validate_communication_file(args.communication_file)


if __name__ == "__main__":
    main()



from __future__ import unicode_literals

import sys
import codecs


def set_stdout_encoding():
    '''
    Force stdout encoding to utf-8.  Ideally the user should set the
    output encoding to utf-8 (or otherwise) in their environment, as
    explained on the internet, but in practice it has been difficult
    to get that right (and scripts writing to stdout have broken).
    '''
    if sys.version_info[0] < 3:
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout)
    else:
        sys.stdout = codecs.getwriter('utf-8')(sys.stdout.detach())



#!/usr/bin/env python

'Convert text file to Concrete Communication file.'
from __future__ import unicode_literals

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import codecs
import logging

import concrete.version
from concrete.util.file_io import write_communication_to_file
from concrete.util.simple_comm import (
    create_comm, AL_NONE, add_annotation_level_argparse_argument
)
from concrete.util import set_stdout_encoding


def main():
    set_stdout_encoding()

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Convert text file to communication',
    )
    parser.set_defaults(annotation_level=AL_NONE)
    parser.add_argument('text_path', type=str,
                        help='Input text file path (- for stdin)')
    parser.add_argument('concrete_path', type=str,
                        help='Output concrete file path (- for stdout)')
    add_annotation_level_argparse_argument(parser)
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    # Won't work on Windows
    text_path = '/dev/fd/0' if args.text_path == '-' else args.text_path
    concrete_path = (
        '/dev/fd/1' if args.concrete_path == '-' else args.concrete_path
    )
    annotation_level = args.annotation_level

    with codecs.open(text_path, encoding='utf-8') as f:
        comm = create_comm(text_path, f.read(),
                           annotation_level=annotation_level)
        write_communication_to_file(comm, concrete_path)


if __name__ == "__main__":
    main()



# This file is NOT auto-generated

"""
Python modules and scripts for working with
`Concrete <https://github.com/hltcoe/concrete>`_, an HLT data
specification defined using `Thrift <http://thrift.apache.org>`_.
"""
from __future__ import absolute_import
from __future__ import unicode_literals


from .access.ttypes import *  # noqa
from .annotate.ttypes import *  # noqa
from .audio.ttypes import *  # noqa
from .clustering.ttypes import *  # noqa
from .communication.ttypes import *  # noqa
from .context.ttypes import *  # noqa
from .email.ttypes import *  # noqa
from .entities.ttypes import *  # noqa
from .exceptions.ttypes import *  # noqa
from .language.ttypes import *  # noqa
from .learn.ttypes import *  # noqa
from .linking.ttypes import *  # noqa
from .metadata.ttypes import *  # noqa
from .nitf.ttypes import *  # noqa
from .search.ttypes import *  # noqa
from .services.ttypes import *  # noqa
from .services.results.ttypes import *  # noqa
from .situations.ttypes import *  # noqa
from .spans.ttypes import *  # noqa
from .structure.ttypes import *  # noqa
from .summarization.ttypes import *  # noqa
from .twitter.ttypes import *  # noqa
from .uuid.ttypes import *  # noqa

from .version import __version__  # noqa



"""Library to validate a Concrete Communication

Validation info, error and warning messages are logged using the
Python standard library's `logging` module.
"""
from __future__ import unicode_literals

import logging

import networkx as nx
from thrift.protocol import TProtocol
from thrift.Thrift import TType

from .util.file_io import read_communication_from_file
from .util.unnone import lun


def validate_communication_file(communication_filename):
    """Test if the :class:`.Communication` in a file is valid

    Deserializes a :class:`.Communication` file into memory, then
    calls :func:`validate_communication` on the Communication object.

    Args:
        communication_filename (str): Name of file containing

    Returns:
        bool
    """
    logging.info(_ilm(
        0, "Opening Concrete Communication with filename '%s'"
           % communication_filename))
    comm = read_communication_from_file(communication_filename)
    return validate_communication(comm)


def validate_communication(comm):
    """Test if all objects in a :class:`.Communication` are valid.

    Calls :func:`validate_thrift_deep` to check for Concrete data
    structure fields that are required by the Concrete Thrift
    definitions.  Then calls:

    - :func:`validate_token_offsets_for_section`
    - :func:`validate_token_offsets_for_sentence`
    - :func:`validate_constituency_parses`
    - :func:`validate_dependency_parses`
    - :func:`validate_token_taggings`
    - :func:`validate_entity_mention_ids`
    - :func:`validate_entity_mention_tokenization_ids`
    - :func:`validate_situations`
    - :func:`validate_situation_mentions`

    Args:
        comm (Communication)

    Returns:
        bool
    """
    valid = True

    logging.info(_ilm(0, "Validating Communication with ID '%s'" % comm.id))

    valid &= validate_thrift_deep(comm)

    for section in lun(comm.sectionList):
        valid &= validate_token_offsets_for_section(section)
        if section.sentenceList:
            logging.debug(_ilm(4, "section '%s' has %d sentences" %
                               (section.uuid, len(section.sentenceList))))
            for sentence in section.sentenceList:
                valid &= validate_token_offsets_for_sentence(sentence)
                if sentence.tokenization:
                    valid &= validate_constituency_parses(
                        comm, sentence.tokenization)
                    valid &= validate_dependency_parses(
                        sentence.tokenization)
                    valid &= validate_token_taggings(sentence.tokenization)

    valid &= validate_entity_mention_ids(comm)
    valid &= validate_entity_mention_tokenization_ids(comm)
    valid &= validate_entity_mention_token_ref_sequences(comm)
    valid &= validate_situations(comm)
    valid &= validate_situation_mentions(comm)

    if not valid:
        logging.error(
            _ilm(0, "The Communication with ID '%s' IS NOT valid" % comm.id))
    else:
        logging.info(
            _ilm(0, "The Communication with ID '%s' is valid" % comm.id))

    return valid


def _get_entity_uuidString_set(comm):
    """
    Args:
        comm (Communication)

    Returns:
        set of strings: uuidStrings for all :class:`.Entity` objects
        in the Communication
    """
    entity_uuidString_set = set()
    for entitySet in lun(comm.entitySetList):
        for entity in lun(entitySet.entityList):
            entity_uuidString_set.add(entity.uuid.uuidString)
    return entity_uuidString_set


def _get_entity_mention_uuidString_set(comm):
    """
    Args:
        comm (Communication)

    Returns:
        set of strings: uuidStrings for all :class:`.EntityMention`
        objects in the Communication
    """
    entity_mention_uuidString_set = set()
    for entityMentionSet in lun(comm.entityMentionSetList):
        for entityMention in lun(entityMentionSet.mentionList):
            entity_mention_uuidString_set.add(
                entityMention.uuid.uuidString)
    return entity_mention_uuidString_set


def _get_sentence_for_tokenization_uuidString_dict(comm):
    """
    Args:
        comm (Communication)

    Returns:
        dict: Maps :class:`.Tokenization` uuidString to :class:`.Sentence`
    """
    if not hasattr(comm, 'sentence_for_tokenization_uuidString_dict'):
        comm.sentence_for_tokenization_uuidString_dict = {}
        for section in lun(comm.sectionList):
            for sentence in lun(section.sentenceList):
                if sentence.tokenization:
                    comm.sentence_for_tokenization_uuidString_dict[
                        sentence.tokenization.uuid.uuidString] = sentence
    return comm.sentence_for_tokenization_uuidString_dict


def _get_situation_uuidString_set(comm):
    """
    Args:
        comm (Communication)

    Returns:
        set of strings: uuidStrings for all :class:`.Situation` objects
        in the :class:`.Communication`
    """
    situation_uuidString_set = set()
    for situationSet in lun(comm.situationSetList):
        for situation in lun(situationSet.situationList):
            situation_uuidString_set.add(situation.uuid.uuidString)
    return situation_uuidString_set


def _get_situation_mention_uuidString_set(comm):
    """
    Args:
        comm (Communication)
    Args:

    - `comm` (`Communication`)

    Returns:

    - set of strings: uuidStrings for all SituationMentions in the
                      Communication
    """
    situation_mention_uuidString_set = set()
    for situationMentionSet in lun(comm.situationMentionSetList):
        for situationMention in lun(situationMentionSet.mentionList):
            situation_mention_uuidString_set.add(
                situationMention.uuid.uuidString)
    return situation_mention_uuidString_set


def _get_tokenization_uuidString_dict(comm):
    """
    Args:
        comm (Communication)
    Args:

    - `comm` (`Communication`)

    Returns:

    - dictionary mapping uuidStrings to Tokenizations
    """
    if not hasattr(comm, '_tokenization_uuidString_dict'):
        comm._tokenization_uuidString_dict = {}
        for section in lun(comm.sectionList):
            for sentence in lun(section.sentenceList):
                tkzn = sentence.tokenization
                if tkzn:
                    u = tkzn.uuid.uuidString
                    comm._tokenization_uuidString_dict[u] = tkzn
    return comm._tokenization_uuidString_dict


def _get_tokenization_uuidString_set(comm):
    """
    Args:
        comm (Communication)
    Args:

    - `comm` (`Communication`)

    Returns:

    - set of strings: uuidStrings for all Tokenizations in the Communication
    """
    tokenization_uuidString_set = set()
    for section in lun(comm.sectionList):
        for sentence in lun(section.sentenceList):
            if sentence.tokenization:
                tokenization_uuidString_set.add(
                    sentence.tokenization.uuid.uuidString)
    return tokenization_uuidString_set


def _ilm(indent_level, log_message):
    """
    ilm = Indented Log Message

    Prepend spaces to a log message so that log messages can be
    printed in a hierarchical list

    Args:

    - `log_message` (string): Log message to be indented
    - `indent_level` (int): Text indentation level

    Returns:

    - string: Indented log message
    """
    return "  " * indent_level + log_message


def validate_constituency_parses(comm, tokenization):
    """Test a :class:`.Tokenization`'s constituency :class:`.Parse`
    objects.

    Verifies that, for each constituent :class:`.Parse`:

    - none of the constituent IDs for the parse repeat
    - the parse tree is a fully connected graph
    - the parse "tree" is really a tree data structure

    Args:
        comm (Communication)
        tokenization (Tokenization)

    Returns:
        bool

    """
    valid = True

    if tokenization.parseList:
        for parse in tokenization.parseList:
            total_constituents = len(parse.constituentList)
            logging.debug(_ilm(6, "tokenization '%s' has %d constituents" % (
                tokenization.uuid, total_constituents)))

            constituent_id_set = set()
            constituent_parse_tree = nx.DiGraph()

            for constituent in parse.constituentList:
                # Add nodes to parse tree
                constituent_parse_tree.add_node(constituent.id)

                if constituent.id not in constituent_id_set:
                    constituent_id_set.add(constituent.id)
                else:
                    valid = False
                    logging.error(_ilm(
                        7, ("constituent ID %d has already been used in this"
                            " sentence's tokenization") % constituent.id))

            # Add edges to constituent parse tree
            for constituent in parse.constituentList:
                if constituent.childList:
                    for child_id in constituent.childList:
                        constituent_parse_tree.add_edge(
                            constituent.id, child_id)

            # Check if constituent parse "tree" is a fully connected graph
            undirected_graph = constituent_parse_tree.to_undirected()
            if not nx.is_connected(undirected_graph):
                valid = False
                logging.error(_ilm(
                    6,
                    "The constituent parse \"tree\" is not a fully"
                    " connected graph - the graph has %d components" %
                    nx.number_connected_components(undirected_graph)))

            # Check if constituent parse "tree" is actually a tree
            if (nx.number_of_nodes(constituent_parse_tree) !=
                    nx.number_of_edges(constituent_parse_tree) + 1):
                valid = False
                logging.error(_ilm(
                    6,
                    "The constituent parse \"tree\" is not a tree."
                    "  |V| != |E|+1  (|V|=%d, |E|=%d)" %
                    (nx.number_of_nodes(constituent_parse_tree),
                     nx.number_of_edges(constituent_parse_tree))))

    return valid


def validate_dependency_parses(tokenization):
    """Test a :class:`.Tokenization`'s :class:`.DependencyParse`
    objects

    Verifies that, for each :class:`.DependencyParse`:

    - the parse is a fully connected graph
    - there are no nodes with a null governer node whose edgeType is not root

    Args:
        tokenization (Tokenization)

    Returns:
        bool
    """
    valid = True

    if tokenization.dependencyParseList:
        total_tokens = len(tokenization.tokenList.tokenList)
        for dependencyParse in tokenization.dependencyParseList:
            dependency_parse_tree = nx.DiGraph()

            # Add nodes to dependency parse tree
            for dependency in dependencyParse.dependencyList:
                if (dependency.gov is None and
                        dependency.edgeType.lower() != "root"):
                    valid = False
                    logging.error(_ilm(
                        7,
                        "Found a null dependency parse node with governer"
                        " whose edgeType is '%s' instead of 'root'" %
                        dependency.edgeType))
                if dependency.gov is not None:
                    if dependency.gov < -1 or dependency.gov > total_tokens:
                        valid = False
                        logging.error(_ilm(
                            7,
                            "Found a null dependency parse node with invalid"
                            " governer of '%d'" %
                            dependency.gov))
                    dependency_parse_tree.add_node(dependency.gov)
                dependency_parse_tree.add_node(dependency.dep)

            # Add edges to dependency parse tree
            for dependency in dependencyParse.dependencyList:
                if dependency.gov is not None:
                    dependency_parse_tree.add_edge(
                        dependency.gov, dependency.dep)

            # Check if dependency parse "tree" is a fully connected graph
            undirected_graph = dependency_parse_tree.to_undirected()
            try:
                if not nx.is_connected(undirected_graph):
                    valid = False
                    logging.error(_ilm(
                        7,
                        ("The dependency parse graph created by '%s' is not"
                         " a fully connected graph - the graph has %d"
                         " components") %
                        (dependencyParse.metadata.tool,
                         nx.number_connected_components(undirected_graph))))
            except nx.exception.NetworkXPointlessConcept:
                logging.warning(_ilm(
                    7,
                    ("The dependency parse graph created by '%s' does not have"
                     " any nodes") % dependencyParse.metadata.tool))
    return valid


def validate_entity_mention_ids(comm):
    """Test if all :class:`.Entity` mentionIds are valid

    Checks if all :class:`.Entity` mentionId :class:`.UUID`'s refer to
    a :class:`.EntityMention` :class:`.UUID` that exists in the
    :class:`.Communication`

    Args:
        comm (Communication)

    Returns:
        bool
    """
    valid = True
    entity_mention_uuidString_set = _get_entity_mention_uuidString_set(comm)

    for entitySet in lun(comm.entitySetList):
        for entity in lun(entitySet.entityList):
            for entityMentionId in entity.mentionIdList:
                if (entityMentionId.uuidString not in
                        entity_mention_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        "Entity '%s' has an invalid entityMentionId (%s)" %
                        (entity.uuid, entityMentionId)))
    return valid


def validate_entity_mention_tokenization_ids(comm):
    """Test `tokenizationID` field of every :class:`.EntityMention`

    Verifies that, for each :class:`.EntityMention`, the
    `entityMention.tokens.tokenizationId` :class:`.UUID` field
    matches the :class:`.UUID` of a  :class:`.Tokenization`
    that exists in this :class:`.Communication`

    Args:
        comm (Communication)

    Returns:
        bool
    """
    valid = True
    tokenization_uuidString_set = _get_tokenization_uuidString_set(comm)

    for entityMentionSet in lun(comm.entityMentionSetList):
        for entityMention in lun(entityMentionSet.mentionList):
            if (entityMention.tokens.tokenizationId.uuidString not in
                    tokenization_uuidString_set):
                valid = False
                logging.error(_ilm(
                    2,
                    "Mention '%s' has an invalid tokenizationId (%s)" %
                    (entityMention.uuid, entityMention.tokens.tokenizationId)))
    return valid


def validate_entity_mention_token_ref_sequences(comm):
    """Test if all :class:`.EntityMention` objects have a valid
    :class:`.TokenRefSequences`

    Args:
        comm (Communication)

    Returns:
        bool

    """
    valid = True
    for entityMentionSet in lun(comm.entityMentionSetList):
        for entityMention in lun(entityMentionSet.mentionList):
            valid &= validate_token_ref_sequence(
                comm, entityMention.tokens)
    return valid


def validate_situation_mentions(comm):
    """Test every :class:`.SituationMention` in the :class:`.Communication`

    A :class:`.SituationMention` has a list of
    :class:`.MentionArgument` objects, and each
    :class:`.MentionArgument` can point to an :class:`.EntityMention`,
    :class:`.SituationMention` or :class:`.TokenRefSequence`.

    Checks that each :class:`.MentionArgument` points to only one type
    of argument.  Also checks validity of all :class:`.EntityMention`
    and :class:`.SituationMention` :class:`.UUID`'s.

    Args:
        comm (Communication)

    Returns:
        bool
    """
    valid = True
    entity_mention_uuidString_set = _get_entity_mention_uuidString_set(comm)
    situation_mention_uuidString_set = _get_situation_mention_uuidString_set(
        comm)

    for situationMentionSet in lun(comm.situationMentionSetList):
        for situationMention in lun(situationMentionSet.mentionList):
            if situationMention.tokens:
                valid &= validate_token_ref_sequence(
                    comm, situationMention.tokens)
            for (m_idx, m_arg) in enumerate(situationMention.argumentList):
                if (m_arg.entityMentionId and
                        m_arg.entityMentionId.uuidString not in
                        entity_mention_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("MentionArgument for SituationMention '%s' has an"
                         " invalid entityMentionId (%s). Tool='%s'") %
                        (situationMention.uuid.uuidString,
                         m_arg.entityMentionId,
                         situationMentionSet.metadata.tool)))
                if (m_arg.situationMentionId and
                        m_arg.situationMentionId.uuidString not in
                        situation_mention_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("MentionArgument for SituationMention '%s' has an"
                         " invalid situationMentionId (%s). Tool='%s'") %
                        (situationMention.uuid,
                         m_arg.situationMentionId,
                         situationMentionSet.metadata.tool)))
                total_args = (
                    bool(m_arg.tokens) +
                    bool(m_arg.entityMentionId) +
                    bool(m_arg.situationMentionId)
                )
                if total_args != 1:
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("MentionArgument #%d for SituationMention '%s'"
                         " should have exactly one EntityMention|"
                         "SituationMention|TokenRefSequence, but found %d") %
                        (m_idx, situationMention.uuid.uuidString,
                         total_args)))
    return valid


def validate_situations(comm):
    """Test every :class:`.Situation` in the :class:`.Communication`

    Checks the validity of all :class:`.EntityMention` and
    :class:`.SituationMention` :class:`.UUID`'s referenced by each
    :class:`.Situation`.

    Args:
        comm (Communication)

    Returns:
        bool

    """
    valid = True

    entity_uuidString_set = _get_entity_uuidString_set(comm)
    situation_mention_uuidString_set = _get_situation_mention_uuidString_set(
        comm)
    situation_uuidString_set = _get_situation_uuidString_set(comm)

    for situationSet in lun(comm.situationSetList):
        for situation in lun(situationSet.situationList):
            for argument in lun(situation.argumentList):
                if (argument.situationId and
                        argument.situationId.uuidString not in
                        situation_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("Argument for Situation '%s' has an invalid"
                         " situationId (%s). Tool='%s'") %
                        (situation.uuid, argument.situationId,
                         situationSet.metadata.tool)))
                if (argument.entityId and
                        argument.entityId.uuidString not in
                        entity_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("Argument for Situation '%s' has an invalid entityId"
                         " (%s). Tool='%s'") %
                        (situation.uuid, argument.entityId,
                         situationSet.metadata.tool)))
            for justification in lun(situation.justificationList):
                if (justification.mentionId.uuidString not in
                        situation_mention_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("Justification for Situation '%s' has an invalid"
                         " [situation] mentionId (%s). Tool='%s'") %
                        (situation.uuid, justification.mentionId,
                         situationSet.metadata.tool)))
                if justification.tokenRefSeqList:
                    for tokenRefSeq in justification.tokenRefSeqList:
                        valid &= validate_token_ref_sequence(
                            comm, tokenRefSeq)
            for mentionId in lun(situation.mentionIdList):
                if (mentionId.uuidString not in
                        situation_mention_uuidString_set):
                    valid = False
                    logging.error(_ilm(
                        2,
                        ("Situation '%s' has an invalid [situation] mentionId"
                         " (%s). Tool='%s'") %
                        (situation.uuid, mentionId,
                         situationSet.metadata.tool)))
    return valid


def validate_token_offsets_for_section(section):
    """Test if the :class:`.TextSpan` boundaries for all
    :class:`.Sentence` objects in a :class:`.Section` fall within the
    boundaries of the :class:`.Section`'s :class:`.TextSpan`

    Args:
        section (Section)

    Returns:
        bool

    """
    valid = True

    if section.textSpan is None:
        return valid

    if section.textSpan.start > section.textSpan.ending:
        valid = False
        logging.error(_ilm(
            2,
            ("Section '%s' has a TextSpan with a start offset (%d) > end"
             " offset (%d)") %
            (section.uuid, section.textSpan.start, section.textSpan.ending)))

    for sentence in lun(section.sentenceList):
        if sentence.textSpan is None:
            continue
        if sentence.textSpan.start > sentence.textSpan.ending:
            valid = False
            logging.error(_ilm(
                2,
                ("Sentence '%s' has a TextSpan with a start offset (%d) > end"
                 " offset (%d)") %
                (sentence.uuid, sentence.textSpan.start,
                 sentence.textSpan.ending)))
        elif ((sentence.textSpan.start < section.textSpan.start) or
                (sentence.textSpan.start > section.textSpan.ending) or
                (sentence.textSpan.ending < section.textSpan.start) or
                (sentence.textSpan.ending > section.textSpan.ending)):
            valid = False
            logging.error(_ilm(
                2,
                ("Sentence '%s' in Section '%s' has a TextSpan [%d, %d] that"
                 " does not fit within the Section TextSpan [%d, %d]") %
                (sentence.uuid, section.uuid, sentence.textSpan.start,
                 sentence.textSpan.ending, section.textSpan.start,
                 section.textSpan.ending)))

    return valid


def validate_token_offsets_for_sentence(sentence):
    """Test if the :class:`.TextSpan` boundaries for all :class:`.Token`
    objects` in a :class:`.Sentence` fall within the boundaries of the
    :class:`.Sentence`'s :class:`.TextSpan`.

    Args:
        sentence (Sentence)

    Returns:
        bool
    """
    valid = True

    if sentence.textSpan is None:
        return valid

    if sentence.textSpan.start > sentence.textSpan.ending:
        valid = False
        logging.error(_ilm(
            7,
            ("Sentence '%s' has a TextSpan with a start offset (%d) > end"
             " offset (%d)") %
            (sentence.uuid, sentence.textSpan.start,
             sentence.textSpan.ending)))
    if sentence.tokenization:
        for token in sentence.tokenization.tokenList.tokenList:
            if token.textSpan is None:
                continue
            if token.textSpan.start > token.textSpan.ending:
                valid = False
                logging.error(_ilm(
                    7,
                    ("Token in Sentence '%s' has a TextSpan with a start"
                     " offset (%d) > end offset (%d)") %
                    (sentence.uuid, token.textSpan.start,
                     token.textSpan.ending)))
            elif ((token.textSpan.start < sentence.textSpan.start) or
                    (token.textSpan.start > sentence.textSpan.ending) or
                    (token.textSpan.ending < sentence.textSpan.start) or
                    (token.textSpan.ending > sentence.textSpan.ending)):
                valid = False
                logging.error(_ilm(
                    7,
                    ("Token in Sentence '%s' has a TextSpan [%d, %d] that does"
                     " not fit within the Sentence TextSpan [%d, %d]") %
                    (sentence.uuid, token.textSpan.start,
                     token.textSpan.ending, sentence.textSpan.start,
                     sentence.textSpan.ending)))

    return valid


def validate_token_ref_sequence(comm, token_ref_sequence):
    """Check if a :class:`.TokenRefSequence` is valid

    Verify that all token indices in the :class:`.TokenRefSequence`
    point to actual token indices in corresponding
    :class:`.Tokenization`

    Args:
        comm (Communication)
        token_ref_sequence (TokenRefSequence)

    Returns:
        bool

    """
    valid = True

    tkzn_map = _get_tokenization_uuidString_dict(comm)
    tkzn_map_sent = _get_sentence_for_tokenization_uuidString_dict(comm)

    if token_ref_sequence.tokenizationId.uuidString not in tkzn_map:
        valid = False
        logging.error(_ilm(
            3,
            "TokenRefSequence has an invalid tokenizationId (%s)" %
            token_ref_sequence.tokenizationId.uuidString))
    else:
        tokenization = tkzn_map[token_ref_sequence.tokenizationId.uuidString]
        for tokenIndex in token_ref_sequence.tokenIndexList:
            try:
                tokenization.tokenList.tokenList[tokenIndex]
            except IndexError:
                valid = False
                logging.error(_ilm(
                    3,
                    "TokenRefSequence '%s' has an invalid tokenIndex (%d)" %
                    (token_ref_sequence.tokenizationId.uuidString,
                     tokenIndex)))
    if token_ref_sequence.tokenizationId.uuidString in tkzn_map_sent:
        sentence = tkzn_map_sent[
            token_ref_sequence.tokenizationId.uuidString]
        if sentence.textSpan and token_ref_sequence.textSpan:
            if ((token_ref_sequence.textSpan.start <
                 sentence.textSpan.start) or
                    (token_ref_sequence.textSpan.start >
                     sentence.textSpan.ending) or
                    (token_ref_sequence.textSpan.ending <
                     sentence.textSpan.start) or
                    (token_ref_sequence.textSpan.ending >
                     sentence.textSpan.ending)):
                valid = False
                logging.error(_ilm(
                    2,
                    ("TokenRefSequence has a TextSpan [%d, %d] that does not"
                     " fit within the Sentence TextSpan [%d, %d]") %
                    (token_ref_sequence.textSpan.start,
                     token_ref_sequence.textSpan.ending,
                     sentence.textSpan.start, sentence.textSpan.ending)))
    return valid


def validate_token_taggings(tokenization):
    """Test if a :class:`.Tokenization` has any :class:`.TokenTagging`
    objects with invalid token indices

    Args:
        tokenization (Tokenization)

    Returns:
        bool
    """
    valid = True
    if tokenization and tokenization.tokenTaggingList:
        total_tokens = len(tokenization.tokenList.tokenList)
        for token_tagging in tokenization.tokenTaggingList:
            for tagged_token in token_tagging.taggedTokenList:
                if (tagged_token.tokenIndex >= total_tokens or
                        tagged_token.tokenIndex < 0):
                    valid = False
                    logging.error(_ilm(
                        7,
                        ("TokenTagging '%s' has a tokenIndex '%d' that is out"
                         " of bounds.") %
                        (token_tagging.uuid.uuidString,
                         tagged_token.tokenIndex)))
    return valid


def validate_thrift(thrift_object, indent_level=0):
    """
    Test if a Thrift object has all required fields.

    This function calls the Thrift object's `validate()` function.
    If an exception is raised because of missing required fields, the
    function catches the exception and logs the exception's error
    message using the Python Standard Library's `logging` module.

    Args:
        thrift_object
        indent_level (int): Text indentation level for logging error message

    Returns:
        bool
    """
    try:
        thrift_object.validate()
    except TProtocol.TProtocolException as e:
        thrift_object_name = str(thrift_object.__class__).split(
            '.')[-1].split("'")[0]
        if hasattr(thrift_object, 'uuid') and thrift_object.uuid is not None:
            thrift_object_name += " '%s'" % thrift_object.uuid
        # For readability, add quotes around field name, changing:
        #   Required field id is unset!
        # to:
        #   Required field 'id' is unset!
        em = e.message.replace("Required field ", "Required Field '").replace(
            " is unset", "' is unset")
        logging.error(_ilm(indent_level, "%s: %s" % (thrift_object_name, em)))
        return False
    else:
        return True


def validate_thrift_object_required_fields(thrift_object, indent_level=0):
    """DEPRECATED: Use :func:`validate_thrift` instead"""
    logging.warning(
        'this rather long name is deprecated and will be removed;'
        ' switch to validate_thrift'
    )
    return validate_thrift(thrift_object, indent_level=indent_level)


def validate_thrift_deep(thrift_object, valid=True):
    """Deep validation of thrift messages.

    Args:
        thrift_object: a Thrift object

    The Python version of Thrift 0.9.1 does not support deep (recursive)
    validation, and none of the Thrift serialization/deserialization
    code calls even the shallow validation functions provided by Thrift.

    This function implements deep validation.  The code is adapted from:

      https://raw.githubusercontent.com/flamholz/py-thrift-validation-example/master/util/validation.py

    See this blog post for more information:

      http://techblog.ridewithvia.com/post/38231652492/recursive-validation-of-python-thrift-structures

    A patch to implement deep validation was submitted to the Thrift
    repository in February of 2013:

      https://issues.apache.org/jira/browse/THRIFT-1732

    but Thrift 0.9.1 - which was released on 2013-08-21 - does not
    include this functionality.
    """
    assert thrift_object is not None
    valid &= validate_thrift(thrift_object)

    # Introspect the structure specification.
    # For each field, check type and decide whether to recurse.
    spec = thrift_object.thrift_spec
    for spec_tuple in spec:
        if spec_tuple is None:
            continue

        mtype = spec_tuple[1]
        name = spec_tuple[2]
        attr = getattr(thrift_object, name)
        if not _ShouldRecurse(mtype):
            # Some primitive type that we don't validate.
            continue

        # Fetch the item itself.
        if attr is None:
            continue

        # Field is set and it's a message or collection, so we validate.
        if mtype == TType.STRUCT:
            valid &= validate_thrift_deep(attr, valid)
        elif mtype in (TType.LIST, TType.SET):
            subtype = spec_tuple[3][0]
            if _ShouldRecurse(subtype):
                for sub_object in attr:
                    valid &= validate_thrift_deep(sub_object, valid)
        elif mtype == TType.MAP:
            subtype = spec_tuple[3]
            key_type = subtype[0]
            val_type = subtype[2]
            for key, val in attr.items():
                if _ShouldRecurse(key_type):
                    valid &= validate_thrift_deep(key, valid)
                if _ShouldRecurse(val_type):
                    valid &= validate_thrift_deep(val, valid)

    return valid


def validate_thrift_object_required_fields_recursively(thrift_object, valid=True):
    """DEPRECATED.  Use :func:`validate_thrift_deep` instead.
    """
    logging.warning(
        'this incredibly long name is deprecated and will be removed;'
        ' switch to validate_thrift_deep'
    )
    return validate_thrift_deep(thrift_object, valid=valid)


_RECURSE_ON = frozenset([TType.STRUCT,
                         TType.LIST,
                         TType.MAP,
                         TType.SET])


def _ShouldRecurse(ttype):
    """Returns True if this ttype is one we recurse on for validation."""
    return ttype in _RECURSE_ON



#!/usr/bin/env python

'Convert tarball of text files to tarball of Concrete communications.'
from __future__ import unicode_literals

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import tarfile
import logging

import concrete.version
from concrete.util.file_io import CommunicationWriterTGZ
from concrete.util.simple_comm import (
    create_comm, AL_NONE, add_annotation_level_argparse_argument
)
from concrete.util import set_stdout_encoding


def load(path, per_line, annotation_level):
    '''
    Generate communications constructed from text files in specified
    tarball, assigning ids that are meaningful tar-friendly filenames.

    If per_line is True:
    One communication is created for each newline in a file.  Note blank
    lines will produce communications.  The trailing newline is included
    in the communication text.  If a file does not have a terminating
    newline, a communication is nonetheless produced for the last line,
    and a newline is appended to the end of the text.
    '''
    with tarfile.open(path, 'r|*') as tf:
        ti = tf.next()
        while ti is not None:
            if ti.isfile():
                f = tf.extractfile(ti)
                text = f.read().decode('utf-8')

                if per_line:
                    if text.endswith('\n'):
                        text = text[:-1]
                    for (i, line) in enumerate(text.split('\n')):
                        yield create_comm(u'%s/%d' % (ti.name, i),
                                          line + u'\n',
                                          annotation_level=annotation_level)
                else:
                    yield create_comm(ti.name, text,
                                      annotation_level=annotation_level)

            tf.members = []
            ti = tf.next()


def main():
    set_stdout_encoding()

    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description='Convert tarball of text files to'
                    ' tarball of concrete communications',
    )
    parser.set_defaults(annotation_level=AL_NONE,
                        log_level='INFO', log_interval=1000)
    parser.add_argument('text_tarball_path', type=str,
                        help='Input text tar file path (- for stdin)')
    parser.add_argument('concrete_tarball_path', type=str,
                        help='Output concrete tar file path (- for stdout)')
    parser.add_argument('--per-line', action='store_true',
                        help='Text files have one document per line (default:'
                             ' each text file is a document)')
    parser.add_argument('--log-interval', type=int,
                        help='Log an info message every log-interval docs')
    add_annotation_level_argparse_argument(parser)
    parser.add_argument('-l', '--loglevel', '--log-level',
                        help='Logging verbosity level threshold (to stderr)',
                        default='info')
    concrete.version.add_argparse_argument(parser)
    args = parser.parse_args()

    # Won't work on Windows
    text_tarball_path = (
        '/dev/fd/0'
        if args.text_tarball_path == '-'
        else args.text_tarball_path
    )
    concrete_tarball_path = (
        '/dev/fd/1'
        if args.concrete_tarball_path == '-'
        else args.concrete_tarball_path
    )
    per_line = args.per_line
    annotation_level = args.annotation_level

    logging.basicConfig(format='%(asctime)-15s %(levelname)s: %(message)s',
                        level=args.loglevel.upper())

    with CommunicationWriterTGZ(concrete_tarball_path) as writer:
        for (i, comm) in enumerate(load(text_tarball_path, per_line,
                                        annotation_level)):
            if (i + 1) % args.log_interval == 0:
                logging.info(u'writing doc %d (%s)...' % (i + 1, comm.id))
            writer.write(comm, comm.id)


if __name__ == "__main__":
    main()
